# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

from pyaoscx.exceptions.generic_op_error import GenericOperationError

from pyaoscx.utils import util as utils

from pyaoscx.configuration import Configuration
from pyaoscx.dns import Dns


class Singleton(type):
    """
    Metaclass to turn classes into a Singleton.
    """

    __instance = None

    def __call__(cls, *args, **kwargs):
        if not cls.__instance:
            cls.__instance = super(Singleton, cls).__call__(*args, **kwargs)
        return cls.__instance


class PyaoscxFactory(metaclass=Singleton):
    """
    Provide a Factory class to instantiate all pyaoscx Modules through specific
        methods. This class is superseded by the Device class, use the Device
        class instead of this one.
    Using the API Version given by the Session.
    """

    def __init__(self, session):
        self.session = session

    def configuration(self):
        """
        Create a Configuration class, to obtain device configuration and
            perform other actions such as backup_config.
        :return: Configuration object.
        """
        config = Configuration(self.session)
        # Get full configuration
        config.get()
        return config

    def dns(
        self,
        vrf=None,
        domain_name=None,
        domain_list=None,
        domain_servers=None,
        host_v4_address_mapping=None,
        host_v6_address_mapping=None,
    ):
        """
        Create a DNS class, to configure a DNS inside a given VRF.
        :param domain_name: Domain name used for name resolution by the DNS
            client, if 'dns_domain_list' is not configured.
        :param domain_list: dict of DNS Domain list names to be used for
            address resolution, keyed by the resolution priority order.
            Example:
                {
                    0: "hpe.com"
                    1: "arubanetworks.com"
                }
        :param domain_servers: dict of DNS Name servers to be used for address
            resolution, keyed by the resolution priority order. Example:
                {
                    0: "4.4.4.10"
                    1: "4.4.4.12"
                }
        :param host_v4_address_mapping: dict of static host address
            configurations and the IPv4 address associated with them. Example:
                {
                    "host1": "5.5.44.5"
                    "host2": "2.2.44.2"
                }
        :param host_v6_address_mapping: dict of static host address
            configurations and the IPv6 address associated with them. Example:
                {
                    "host1": "2001:0db8:85a3:0000:0000:8a2e:0370:7334"
                }
        : return DNS object.
        """
        if vrf is None:
            vrf = "default"

        if domain_list is None:
            domain_list = {}
        if domain_servers is None:
            domain_servers = {}
        if host_v4_address_mapping is None:
            host_v4_address_mapping = {}
        if host_v6_address_mapping is None:
            host_v6_address_mapping = {}

        # Ensure that all keys are integers
        domain_list = {int(k): v for k, v in domain_list.items()}
        domain_servers = {int(k): v for k, v in domain_servers.items()}

        # Create DNS object
        dns = Dns(
            self.session,
            vrf,
            domain_name,
            domain_list,
            domain_servers,
            host_v4_address_mapping,
            host_v6_address_mapping,
        )

        # Apply object into Switch
        dns.apply()

        return dns

    def interface(self, name):
        """
        Create an Interface object.
        :param name: Alphanumeric name of Interface.
        :return: Interface object.
        """
        interface_obj = self.session.api.get_module(
            self.session, "Interface", name
        )
        try:
            # Try to create the interface
            interface_obj.create()

        except GenericOperationError:
            # The GenericOperationError is risen if the POST request was
            # executed correctly, but the switch didn't accept it. This means
            # that the interface already exists.
            interface_obj.get()

        return interface_obj

    def ipv6(self, address, interface_name, address_type=None):
        """
        Create a Ipv6 object. If values differ from existing object, incoming
            changes will be applied.
        :param address: Alphanumeric address of IPv6. Example:
            '2001:db8::11/ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff'
        :param interface_name: Alphanumeric name of the Interface parent of the
            IPv6 Address. A Interface object is also accepted.
        :param address_type: Type of IPv6 address. Defaults to "global-unicast"
            if not specified..
        :return: Ipv6 object.
        """
        if address_type is None:
            _type = "global-unicast"
        else:
            _type = address_type

        if isinstance(interface_name, str):
            # Make Interface into an object
            interface = self.session.api.get_module(
                self.session, "Interface", interface_name
            )
            # Materialize interface to ensure its existence
            interface.get()

        ipv6_obj = self.session.api.get_module(
            self.session,
            "Ipv6",
            address,
            parent_int=interface,
            type=_type,
            preferred_lifetime=604800,
            valid_lifetime=2592000,
            node_address=True,
            ra_prefix=True,
            route_tag=0,
        )

        # Try to obtain data; if not, create
        try:
            ipv6_obj.get()
            # Configure variables in case something changes
            if address_type is not None:
                ipv6_obj.type = address_type
            ipv6_obj.apply()
        except GenericOperationError:
            # Create object inside switch
            ipv6_obj.apply()

        return ipv6_obj

    def vlan(
        self,
        vlan_id,
        name=None,
        description=None,
        vlan_type=None,
        admin_conf_state="up",
    ):
        """
        Create a Vlan object.
        :param vlan_id: Numeric ID for VLAN.
        :param name: Alphanumeric name of VLAN, Defaults to "VLAN <ID>".
        :param description: Optional description to add to VLAN.
        :param vlan_type: VLAN type. Defaults to "static" if not specified.
        :param admin_conf_state: Optional administratively-configured state of
            VLAN. Only configurable for static VLANs. Defaults to "up" for
            static VLANs.
        :return: Vlan object.
        """
        if name is None:
            name = "VLAN {0}".format(str(vlan_id))

        if vlan_type is None:
            pvlan_type = "static"
        else:
            pvlan_type = vlan_type

        if pvlan_type == "static":
            # admin-configured state can only be set on static VLANs
            vlan_obj = self.session.api.get_module(
                self.session,
                "Vlan",
                vlan_id,
                name=name,
                description=description,
                admin=admin_conf_state,
            )
        else:
            vlan_obj = self.session.api.get_module(
                self.session,
                "Vlan",
                vlan_id,
                name=name,
                description=description,
            )

        # Try to obtain data; if not, create
        try:
            vlan_obj.get()
            # Configure variables in case something changes
            if name is not None:
                vlan_obj.name = name
            if description is not None:
                vlan_obj.description = description
            if admin_conf_state is not None and vlan_type == "static":
                vlan_obj.admin = admin_conf_state
            vlan_obj.apply()

        except Exception:
            # Create object inside switch
            vlan_obj.apply()

        return vlan_obj

    def __get_vrf_from_switch(self, name):
        """
        Get VRF from switch, this avoids making changes to the VRF object. Note
            the __double_leading_underscore that signifies that this is meant
            to be used only inside this module.
        :param name: VRF name.
        :return: Materialized Vrf object.
        """
        vrf = self.session.api.get_module(self.session, "Vrf", name)
        vrf.get()
        return vrf

    def vrf(self, name, route_distinguisher=None, vrf_type=None):
        """
        Create a Vrf object. If values differ from existing object, incoming
            changes will be applied.
        :param name: VRF name.
        :param route_distinguisher: Optional route distinguisher to add.
            Defaults to nothing if not specified.
        :param vrf_type: Optional VRF type. Defaults to "user" if not
            specified.
        :return: Vrf object.
        """
        if vrf_type is None:
            _type = "user"
        else:
            _type = vrf_type

        if route_distinguisher is not None and _type != "default":
            vrf_obj = self.session.api.get_module(
                self.session, "Vrf", name, rd=route_distinguisher, type=_type
            )
        else:
            vrf_obj = self.session.api.get_module(
                self.session, "Vrf", name, type=_type
            )

        # Try to obtain data; if not, create
        try:
            vrf_obj.get()
            # Configure variables in case something changes
            if route_distinguisher is not None:
                vrf_obj.rd = route_distinguisher
            # Apply changes wanted by user
            vrf_obj.apply()
        except GenericOperationError:
            # Create object inside switch
            vrf_obj.apply()

        return vrf_obj

    def vsx(self, **kwargs):
        """
        Create a Vsx object.  If values differ from existing object, incoming
            changes will be applied.
        :return: A Vsx object.
        :rtype: Vsx.
        """
        keepalive_vrf = kwargs.get("keepalive_vrf")
        if keepalive_vrf:
            if isinstance(keepalive_vrf, str):
                keepalive_vrf = self.__get_vrf_from_switch(keepalive_vrf)
            kwargs["keepalive_vrf"] = keepalive_vrf.get_info_format()
        software_update_vrf = kwargs.get("software_update_vrf")
        if software_update_vrf:
            if isinstance(software_update_vrf, str):
                software_update_vrf = self.__get_vrf_from_switch(
                    software_update_vrf
                )
            kwargs[
                "software_update_vrf"
            ] = software_update_vrf.get_info_format()
        isl_port = kwargs.get("isl_port")
        if isl_port:
            if isinstance(isl_port, str):
                isl_port = self.session.api.get_module(
                    self.session, "Interface", isl_port
                )
                isl_port.get()
            # Check ISL Port routing
            if isl_port.routing:
                # Set routing to False
                isl_port.routing = False
                isl_port.apply()
            kwargs["isl_port"] = isl_port.get_info_format()
        vsx_obj = self.session.api.get_module(self.session, "Vsx", **kwargs)
        # Try to obtain data; if not, create
        try:
            vsx_obj.get()
            # get() overwrites object's attributes with the switch config, so
            # set them here after object is materialized, to keep the new ones
            for key, value in kwargs.items():
                setattr(vsx_obj, key, value)
        except GenericOperationError:
            pass
        finally:
            # Create object inside switch
            vsx_obj.apply()
        return vsx_obj

    def bgp_router_asn(self, vrf, asn, router_id=None):
        """
        Create a BgpRouter object as Autonomous System Number. If values differ
            from existing object, incoming changes will be applied.
        :param vrf: Alphanumeric name of the VRF the BGP ASN belongs to. A Vrf
            object is also accepted.
        :param asn: Integer that represents the Autonomous System Number.
        :param router_id: Optional IPv4 address that functions as the BGP
            Router ID.
        :return: BgpRouter object.
        """
        if isinstance(vrf, str):
            vrf = self.__get_vrf_from_switch(vrf)

        bgp_router_obj = self.session.api.get_module(
            self.session, "BgpRouter", asn, parent_vrf=vrf, router_id=router_id
        )

        # Try to obtain data; if not, create
        try:
            bgp_router_obj.get()
            # Change attributes
            if router_id is not None:
                bgp_router_obj.router_id = router_id
            # Apply changes
            bgp_router_obj.apply()
        except GenericOperationError:
            # Create object inside switch
            bgp_router_obj.apply()

        return bgp_router_obj

    def bgp_router_vrf(self, vrf, asn, redistribute):
        """
        Create a BgpRouter object with a BGP VRF settings for the associated
            BGP ASN. If values differ from existing object, incoming changes
            will be applied.
        :param vrf: Alphanumeric name of the VRF the BGP ASN belongs to. A Vrf
            object is also accepted.
        :param asn: Integer that represents the Autonomous System Number.
        :param redistribute: Alphanumeric to specify which types of routes that
            should be redistributed by BGP. The options are "ipv4-unicast" or
            "ipv6-unicast".
        :return: BgpRouter object.
        """
        if isinstance(vrf, str):
            vrf = self.__get_vrf_from_switch(vrf)

        redistribute_data = {}

        if redistribute == "ipv4-unicast":
            redistribute_data = {"ipv4-unicast": ["connected"]}
        elif redistribute == "ipv6-unicast":
            redistribute_data = {"ipv6-unicast": ["connected"]}

        bgp_router_obj = self.session.api.get_module(
            self.session,
            "BgpRouter",
            asn,
            parent_vrf=vrf,
            redistribute=redistribute_data,
        )

        # Try to obtain data; if not, create
        try:
            bgp_router_obj.get()
            # Change attributes
            bgp_router_obj.redistribute = redistribute_data
            # Apply changes
            bgp_router_obj.apply()
        except GenericOperationError:
            # Create object inside switch
            bgp_router_obj.apply()

        return bgp_router_obj

    def bgp_neighbor(
        self,
        vrf,
        bgp_router_asn,
        group_ip,
        family_type=None,
        reflector=None,
        send_community=None,
        local_interface="",
    ):
        """
        Create a BgpNeighbor object. If values differ from existing object,
            incoming changes will be applied.
        :param vrf: Alphanumeric name of the VRF the BGP ASN belongs to.
            A Vrf object is also accepted.
        :param bgp_router_asn: Integer that represents the Autonomous System
            Number.
        :param group_ip: IPv4 address or name of group of the neighbors that
            functions as the BGP Router link. Example: '1.1.1.1'
        :param family_type: Alphanumeric to specify what type of neighbor
            settings to configure. The options are 'l2vpn-evpn',
            'ipv4-unicast', or 'ipv6-unicast'. When setting to l2vpn-evpn,
            the neighbor configurations also will add route-reflector-client
            and send-community settings.
        :param reflector: Boolean value to determine whether this neighbor
            has route reflector enabled. Default is False.
        :param send_community: Boolean value to determine whether this neighbor
            has send-community enabled. Default is False.
        :param local_interface: Optional alphanumeric to specify which
            interface the neighbor will apply to.
        :return: BgpNeighbor object.
        """
        if family_type is None:
            _family_type = "l2vpn-evpn"

        _valid_family_types = ["l2vpn-evpn", "ipv4-unicast", "ipv6-unicast"]
        if _family_type not in _valid_family_types:
            raise Exception(
                "ERROR: family_type must be one of: {0}".format(
                    _valid_family_types
                )
            )

        if isinstance(vrf, str):
            vrf = self.__get_vrf_from_switch(vrf)

        if isinstance(bgp_router_asn, int):
            # Make BGP Router into an object
            bgp_router_obj = self.session.api.get_module(
                self.session, "BgpRouter", bgp_router_asn, parent_vrf=vrf
            )

            # Materialize BGP Router to ensure its existence
            bgp_router_obj.get()
            # Set asn integer
            asn = bgp_router_asn
            # Set variable as an object
            bgp_router_asn = bgp_router_obj
        else:
            asn = bgp_router_asn.asn

        if local_interface != "":
            if isinstance(local_interface, str):
                local_interface = self.session.api.get_module(
                    self.session, "Interface", local_interface
                )
                local_interface.get()

        # Set values needed
        activate = {
            "ipv4-unicast": False,
            "ipv6-unicast": False,
            "l2vpn-evpn": False,
        }

        next_hop_unchanged = {"l2vpn-evpn": False}

        route_reflector_client = {
            "ipv4-unicast": False,
            "ipv6-unicast": False,
            "l2vpn-evpn": False,
        }

        send_community_data = {
            "ipv4-unicast": "none",
            "ipv6-unicast": "none",
            "l2vpn-evpn": "none",
        }

        activate[_family_type] = True

        # Set incoming variables
        if send_community is None:
            _send_community = False
        else:
            _send_community = send_community
        if reflector is None:
            _reflector = False
        else:
            _reflector = reflector

        if _send_community:
            send_community_data[_family_type] = "both"

        if _reflector:
            route_reflector_client[_family_type] = reflector

        bgp_neighbor_obj = self.session.api.get_module(
            self.session,
            "BgpNeighbor",
            group_ip,
            parent_bgp_router=bgp_router_asn,
            remote_as=asn,
            shutdown=False,
            local_interface=local_interface,
            activate=activate,
            next_hop_unchanged=next_hop_unchanged,
            route_reflector_client=route_reflector_client,
            send_community=send_community_data,
        )

        # Try to obtain data; if not, create
        try:
            bgp_neighbor_obj.get()
            # Change attributes
            if local_interface != "":
                bgp_neighbor_obj.local_interface = local_interface
            if family_type is not None:
                bgp_neighbor_obj.activate = activate
            if send_community is not None:
                bgp_neighbor_obj.send_community = send_community_data
            if reflector is not None:
                bgp_neighbor_obj.route_reflector_client = (
                    route_reflector_client
                )
            # Apply changes
            bgp_neighbor_obj.apply()
        except GenericOperationError:
            # Create object inside switch
            bgp_neighbor_obj.apply()

        return bgp_neighbor_obj

    def ospf_router(self, vrf, ospf_id, **kwargs):
        """
        Create OspfRouter object. If values differ from existing object,
            incoming changes will be applied.
        :param vrf: Alphanumeric name of the VRF the OSPF ID belongs to A Vrf
            object is also accepted.
        :param ospf_id: OSPF process ID between numbers 1-63.
        :param redistribute: List of types of redistribution methods for the
            OSPF Process, with the options being "bgp", "connected", and
            "static".
        :return: OspfRouter object.
        """
        if "redistribute" not in kwargs:
            kwargs["redistribute"] = ["connected", "static"]
        if isinstance(vrf, str):
            vrf = self.__get_vrf_from_switch(vrf)
        ospf_router_obj = self.session.api.get_module(
            self.session, "OspfRouter", ospf_id, parent_vrf=vrf, **kwargs
        )
        # Try to obtain data; if not, create
        try:
            ospf_router_obj.get()
            # get() overwrites object's attributes with the switch config, so
            # set them here after object is materialized, to keep the new ones
            utils.set_config_attrs(ospf_router_obj, kwargs)
            ospf_router_obj.apply()
        except GenericOperationError:
            ospf_router_obj.create()
        return ospf_router_obj

    def ospfv3_router(self, vrf, ospfv3_id, **kwargs):
        """
        Create Ospfv3Router object. If values differ from existing object,
            incoming changes will be applied.
        :param vrf: Alphanumeric name of the VRF the OSPF ID belongs to.
            A Vrf object is also accepted.
        :param ospfv3_id: OSPF process ID between numbers 1-63.
        :param redistribute: List of types of redistribution methods for the
        OSPF Process, with the options being "bgp", "connected",
        "local_loopback", "rip", and "static".
        :return: Ospfv3Router object.
        """
        if "redistribute" not in kwargs:
            kwargs["redistribute"] = ["connected", "static"]
        if isinstance(vrf, str):
            vrf = self.__get_vrf_from_switch(vrf)
        ospfv3_router_obj = self.session.api.get_module(
            self.session, "Ospfv3Router", ospfv3_id, parent_vrf=vrf, **kwargs
        )
        # Try to obtain data; if not, create
        try:
            ospfv3_router_obj.get()
            # get() overwrites object's attributes with the switch config, so
            # set them here after object is materialized, to keep the new ones
            utils.set_config_attrs(ospfv3_router_obj, kwargs)
            ospfv3_router_obj.apply()
        except GenericOperationError:
            ospfv3_router_obj.create()
        return ospfv3_router_obj

    def ospf_router_area(self, vrf, ospf_router, area_id, **kwargs):
        """
        Create an OspfArea object. If values differ from existing object,
            incoming changes will be applied.
        :param vrf: Alphanumeric name of the VRF the OSPF ID belongs to.
        :param ospf_router: OSPF process ID in [1, 63], an OspfRouter, or
            Ospfv3Router object is also accepted.
        :param area_id: Unique identifier as a string in the form of x.x.x.x.
        :param area_type: Alphanumeric defining how the external routing and
            summary LSAs for this area will be handled. Options are "default",
            "nssa", "nssa_no_summary", "stub", "stub_no_summary". If no value
            is passed, "default" is used.
        :return: OspfArea object.
        """
        if "area_type" not in kwargs:
            kwargs["area_type"] = "default"
        if "ipsec_ah" not in kwargs:
            kwargs["ipsec_ah"] = {}
        if "ipsec_esp" not in kwargs:
            kwargs["ipsec_esp"] = {}
        if isinstance(vrf, str):
            vrf = self.__get_vrf_from_switch(vrf)
        router = ospf_router
        if isinstance(router, int):
            # Make OSPF Router into an object
            ospf_router_obj = self.session.api.get_module(
                self.session, "OspfRouter", ospf_router, parent_vrf=vrf
            )
            # Materialize OSPF Router to ensure its existence
            ospf_router_obj.get()
            # Set variable as an object
            router = ospf_router_obj
        # Create OspfArea object
        ospf_area_obj = self.session.api.get_module(
            self.session,
            "OspfArea",
            area_id,
            parent_ospf_router=router,
            **kwargs
        )
        # Try to obtain data; if not, create
        try:
            ospf_area_obj.get()
            # get() overwrites object's attributes with the switch config, so
            # set them here after object is materialized, to keep the new ones
            utils.set_config_attrs(ospf_area_obj, kwargs)
            ospf_area_obj.apply()
        except GenericOperationError:
            ospf_area_obj.create()
        return ospf_area_obj

    def __get_ospf_router(self, ospf_id, vrf):
        """
        Get OSPF Router object from switch, this avoids making changes to it.
            Note the __double_leading_underscore that signifies that this is
            meant to be used only inside this module.
        :param ospf_id: OSPF process ID in [1,63].
        :param vrf: Vrf (object).
        :return: Materialized OspfRouter (object).
        """
        ospf_router_obj = self.session.api.get_module(
            self.session, "OspfRouter", ospf_id, parent_vrf=vrf
        )
        ospf_router_obj.get()
        return ospf_router_obj

    def __get_ospfv3_router(self, ospfv3_id, vrf):
        """
        Get OSPFv3 Router object from switch, this avoids making changes to it.
            Note the __double_leading_underscore that signifies that this is
            meant to be used only inside this module.
        :param ospfv3_id: OSPFv3 process ID in [1,63].
        :param vrf: Vrf (object).
        :return: Materialized Ospfv3Router (object).
        """
        ospfv3_router_obj = self.session.api.get_module(
            self.session, "Ospfv3Router", ospfv3_id, parent_vrf=vrf
        )
        ospfv3_router_obj.get()
        return ospfv3_router_obj

    def __get_ospf_area(self, area_id, ospf_router):
        """
        Get OSPFv3 Area object from switch, this avoids making changes to it.
            Note the __double_leading_underscore that signifies that this is
            meant to be used only inside this module.
        :param area_id: Unique identifier as a string in the form of x.x.x.x.
        :param ospf_router: OspfRouter (object) or Ospfv3Router (object).
        :return: Materialized Ospfv3Router (object).
        """
        ospf_area_obj = self.session.api.get_module(
            self.session, "OspfArea", area_id, parent_ospf_router=ospf_router
        )
        ospf_area_obj.get()
        return ospf_area_obj

    def ospfv3_router_area(self, vrf, ospfv3_id, area_id, **kwargs):
        """
        Create OspfArea object. If values differ from existing object, incoming
            changes will be applied.
        :param vrf: Alphanumeric name of the VRF the OSPF ID belongs to.
        :param ospfv3_id: OSPFv3 process ID in [1,63], an Ospfv3Router object
            is also accepted.
        :param area_id: Unique identifier as a string in the form of x.x.x.x.
        :param area_type: Alphanumeric defining how the external routing and
            summary LSAs for this area will be handled. Options are "default",
            "nssa", "nssa_no_summary", "stub", "stub_no_summary" if no value
            is passed, "default" is used.
        :return: OspfArea object.
        """
        if isinstance(vrf, str):
            vrf = self.__get_vrf_from_switch(vrf)
        ospfv3_router = ospfv3_id
        if isinstance(ospfv3_id, int):
            ospfv3_router = self.__get_ospfv3_router(ospfv3_id, vrf)
        return self.ospf_router_area(vrf, ospfv3_router, area_id, **kwargs)

    def ospf_interface(self, vrf, ospf_id, area_id, interface_name, **kwargs):
        """
        Create a OspfInterface object.
        :param vrf: Alphanumeric name of the VRF the OSPF ID belongs to. A Vrf
            object is also accepted.
        :param ospf_id: OSPF process ID between numbers 1-63. A OSPF Router is
            accepted.
        :param area_id: Unique identifier as a string in the form of x.x.x.x.
        :param interface_name: Alphanumeric name of the interface that will be
            attached to the OSPF area.
        :return: OspfInterface object.
        """
        if isinstance(vrf, str):
            vrf = self.__get_vrf_from_switch(vrf)
        ospf_router = ospf_id
        if isinstance(ospf_id, int):
            ospf_router = self.__get_ospf_router(ospf_id, vrf)
        area = area_id
        if isinstance(area_id, str):
            area = self.__get_ospf_area(area_id, ospf_router)
        # Make Ospf ID into an object
        ospf_interface = self.session.api.get_module(
            self.session,
            "OspfInterface",
            interface_name,
            parent_ospf_area=area,
            **kwargs
        )
        # Try to obtain data; if not, create
        try:
            ospf_interface.get()
        except GenericOperationError:
            # Create object inside switch
            ospf_interface.apply()

        return ospf_interface

    def ospfv3_interface(
        self, vrf, ospf_id, area_id, interface_name, **kwargs
    ):
        """
        Create a OspfInterface object.
        :param vrf: Alphanumeric name of the VRF the OSPF ID belongs to. A Vrf
            object is also accepted.
        :param ospfv3_id: OSPFv3 process ID between numbers 1-63. An OSPFv3
            Router is accepted.
        :param area_id: Unique identifier as a string in the form of x.x.x.x.
        :param interface_name: Alphanumeric name of the interface that will be
            attached to the OSPF area.
        :return: OspfInterface object.
        """
        if isinstance(vrf, str):
            vrf = self.__get_vrf_from_switch(vrf)
        ospf_router = ospf_id
        if isinstance(ospf_id, int):
            ospf_router = self.__get_ospfv3_router(ospf_id, vrf)
        return self.ospf_interface(
            vrf, ospf_router, area_id, interface_name, **kwargs
        )

    def ospf_vlink(self, vrf, ospf_router, area_id, peer_router, **kwargs):
        """
        Create a OspfVlink object. Defaults to using OspfRouter, when providing
            the OSPF process ID.
        :param vrf: Alphanumeric name of the VRF the OSPF ID belongs to. A Vrf
            object is also accepted.
        :param ospf_router: OSPF process ID number in [1,63]. An OSPF Router
            object is also accepted (both v2, and v3 versions).
        :param area_id: Unique identifier as a string in the form of x.x.x.x.
        :param peer_router: ID of the peer OSPF router as IP address in x.x.x.x
            form.
        :return: OspfVlink object.
        """
        if isinstance(vrf, str):
            vrf = self.__get_vrf_from_switch(vrf)
        if isinstance(ospf_router, int):
            router = self.__get_ospf_router(ospf_router, vrf)
        area = area_id
        if isinstance(area_id, str):
            area = self.__get_ospf_area(area_id, router)
        # Get OspfVlink from peer_router
        vlink = self.session.api.get_module(
            self.session,
            "OspfVlink",
            peer_router,
            parent_ospf_area=area,
            **kwargs
        )
        try:
            # Get the remote configuration, but the local one takes precedence
            vlink.get()
            utils.set_config_attrs(vlink, kwargs)  # so it gets re-applied here
            vlink.apply()
        except GenericOperationError:
            vlink.apply()  # if can't get it, create
        return vlink

    def ospfv3_vlink(self, vrf, ospf_router, area_id, peer_router, **kwargs):
        """
        Create a OspfVlink object.
        :param vrf: Alphanumeric name of the VRF the OSPF ID belongs to. A Vrf
            object is also accepted.
        :param ospf_router: OSPFv3 process ID number in [1, 63]. An
             Ospfv3Router object is also accepted.
        :param area_id: Unique identifier as a string in the form of x.x.x.x.
        :param peer_router: ID of the peer OSPFv3 router as IP address in
            x.x.x.x form.
        :return: OspfVlink object.
        """
        if isinstance(vrf, str):
            vrf = self.__get_vrf_from_switch(vrf)
        if isinstance(ospf_router, int):
            router = self.__get_ospfv3_router(ospf_router, vrf)
        return self.ospf_vlink(vrf, router, area_id, peer_router, **kwargs)

    def vlan_and_svi(
        self,
        vlan_id,
        vlan_name,
        vlan_int_name,
        vlan_desc=None,
        ipv4=None,
        vrf_name="default",
        vlan_port_desc=None,
    ):
        """
        Create VLAN and Interface objects to represent VLAN and SVI,
            respectively.
        :param vlan_id: Numeric ID of VLAN.
        :param vlan_name: Alphanumeric name of VLAN.
        :param vlan_int_name: Alphanumeric name for the VLAN interface.
        :param vlan_desc: Optional description to add to VLAN.
        :param ipv4: Optional IPv4 address to assign to the interface. Defaults
            to nothing if not specified. Example: '1.1.1.1'
        :param vrf_name: VRF to attach the SVI to. Defaults to "default".
        :param vlan_port_desc: Optional description for the interface.
            Defaults to nothing if not specified.
        :return: A tuple with a Vlan object and a Interface SVI object.
        """
        # Create Vlan object
        vlan_obj = self.vlan(vlan_id, vlan_name, vlan_desc)

        # Create Interface Object
        interface_obj = self.interface(vlan_int_name)
        # Set Interface as an SVI
        interface_obj.configure_svi(vlan_id, ipv4, vrf_name, vlan_port_desc)

        return vlan_obj, interface_obj

    def dhcp_relay(self, vrf, port):
        """
        Create a DhcpRelay object.
        :param vrf: Alphanumeric name of VRF.
        :param port: Alphanumeric name of Port.
        :return: DhcpRelay object.
        """
        port_obj = self.session.api.get_module(self.session, "Interface", port)
        vrf_obj = self.session.api.get_module(self.session, "Vrf", vrf)

        dhcp_relay = self.session.api.get_module(
            self.session, "DhcpRelay", index_id=vrf_obj, port=port_obj
        )

        # Try to obtain data; if not, create
        try:
            dhcp_relay.get()
        except GenericOperationError:
            # Create object inside switch
            dhcp_relay.apply()

        return dhcp_relay

    def acl(self, list_name, list_type):
        """
        Create an Acl object.
        :param list_name: Alphanumeric name of ACL.
        :param list_type: Alphanumeric type of ACL. Type should be one of
            "ipv4," "ipv6," or "mac".
        :return: Acl object.
        """
        acl = self.session.api.get_module(
            self.session, "ACL", index_id=list_name, list_type=list_type
        )

        # Try to obtain data; if not, create
        try:
            acl.get()
        except GenericOperationError:
            # Create object inside switch
            acl.apply()

        return acl

    def acl_entry(
        self,
        list_name,
        list_type,
        sequence_num,
        action="permit",
        count=None,
        protocol=None,
        src_ip=None,
        dst_ip=None,
        dst_l4_port_min=None,
        dst_l4_port_max=None,
        src_mac=None,
        dst_mac=None,
        ethertype=None,
        **kwargs
    ):
        """
        Create an AclEntry object.
        :param list_name: Alphanumeric name of the ACL.
        :param list_type: Type should be one of "ipv4," "ipv6," or "mac".
        :param sequence_num: Integer number of the sequence.
        :param action: Action should be either "permit" or "deny".
        :param count: Optional boolean flag that when true, will make entry
            increment hit count for matched packets.
        :param protocol: Optional integer IP protocol number.
        :param src_ip: Optional source IP address. Both IPv4 and IPv6 are
            supported. Example:
                10.10.12.11/255.255.255.255
                2001:db8::11/ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff
        :param dst_ip: Optional destination IP address. Both IPv4 and IPv6 are
            supported. Example:
                10.10.12.11/255.255.255.255
                2001:db8::11/ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff
        :param dst_l4_port_min: Optional minimum L4 port number in range; used
            in conjunction with dst_l4_port_max.
        :param dst_l4_port_max: Optional maximum L4 port number in range; used
            in conjunction with dst_l4_port_min.
        :param src_mac: Optional source MAC address. Example:
            '01:02:03:04:05:06'
        :param dst_mac: Optional destination MAC address. Example:
            '01:02:03:04:05:06'
        :param ethertype: Optional integer EtherType number.
        :param kwargs: Optional keyword arguments for more detailed
            configuration.
        :return acl_entry: A AclEntry object.
        """
        # Create Acl object
        acl = self.session.api.get_module(
            self.session, "ACL", index_id=list_name, list_type=list_type
        )

        # Get ACL data
        acl.get()

        # Create ACL Entry
        acl_entry_obj = self.session.api.get_module(
            self.session,
            "AclEntry",
            index_id=sequence_num,
            parent_acl=acl,
            action=action,
            count=count,
            protocol=protocol,
            src_ip=src_ip,
            dst_ip=dst_ip,
            dst_l4_port_min=dst_l4_port_min,
            dst_l4_port_max=dst_l4_port_max,
            src_mac=src_mac,
            dst_mac=dst_mac,
            ethertype=ethertype,
            **kwargs
        )

        # Try to obtain data; if not, create
        try:
            acl_entry_obj.get()
            # Change attributes
            if dst_l4_port_min is not None:
                acl_entry_obj.dst_l4_port_min = dst_l4_port_min
            if dst_l4_port_max is not None:
                acl_entry_obj.dst_l4_port_max = dst_l4_port_max
            if src_mac is not None:
                acl_entry_obj.src_mac = src_mac
            if dst_mac is not None:
                acl_entry_obj.dst_mac = dst_mac
            if ethertype is not None:
                acl_entry_obj.ethertype = ethertype
            # Apply changes
            acl_entry_obj.apply()
        except GenericOperationError:
            # Create object inside switch
            acl_entry_obj.apply()

        return acl_entry_obj

    def vrf_address_family(self, vrf, address_family="ipv4_unicast"):
        """
        Create a VrfAddressFamily object with a VRF.
        :param vrf: Alphanumeric name of the VRF the Family Address belongs to.
            A Vrf object is also accepted.
        :param address_family: Alphanumeric type of the Address Family. The
            options are 'ipv4_unicast' and 'ipv6_unicast'. Defaults to
            'ipv4_unicast'.
        :return: VRF_Address_Family object.
        """
        if isinstance(vrf, str):
            vrf = self.__get_vrf_from_switch(vrf)

        vrf_address_fam_obj = self.session.api.get_module(
            self.session, "VrfAddressFamily", address_family, parent_vrf=vrf
        )

        # Try to obtain data; if not, create
        try:
            vrf_address_fam_obj.get()
        except GenericOperationError:
            # Create object inside switch
            vrf_address_fam_obj.apply()

        return vrf_address_fam_obj

    def aggregate_address(self, vrf, bgp_router_asn, family_type, ip_prefix):
        """
        Create an AggregateAddress object.
        :param vrf: Alphanumeric name of the VRF the BGP ASN belongs to. A Vrf
            object is also accepted.
        :param bgp_router_asn: Integer that represents the Autonomous System
            Number.
        :param family_type: Address Family type for the Aggregate Address.
            The options are: 'ipv4-unicast', 'ipv6-unicast'.
        :param ip_prefix: IP address and mask used to key Aggregate Address.
            Example: '1.1.1.1/24'.
        :return: AggregateAddress object.
        """
        _valid_family_types = ["ipv4-unicast", "ipv6-unicast"]
        if family_type not in _valid_family_types:
            raise Exception(
                "ERROR: family_type must be one of: {0}".format(
                    _valid_family_types
                )
            )

        if isinstance(vrf, str):
            vrf = self.__get_vrf_from_switch(vrf)

        if isinstance(bgp_router_asn, int):
            # Make BGP Router into an object
            bgp_router_obj = self.session.api.get_module(
                self.session, "BgpRouter", bgp_router_asn, parent_vrf=vrf
            )

            # Materialize interface to ensure its existence
            bgp_router_obj.get()
            # Set variable as an object
            bgp_router_asn = bgp_router_obj

        aggregate_add_obj = self.session.api.get_module(
            self.session,
            "AggregateAddress",
            family_type,
            ip_prefix=ip_prefix,
            parent_bgp_router=bgp_router_asn,
        )

        # Try to obtain data; if not, create
        try:
            aggregate_add_obj.get()
        except GenericOperationError:
            # Create object inside switch
            aggregate_add_obj.apply()

        return aggregate_add_obj

    def static_route(self, vrf, destination_address_prefix):
        """
        Create a StaticRoute object with a VRF.
        :param vrf: Name of the VRF on which the static route is to be
            configured. Defaults to default vrf. A Vrf object is also accepted.
        :param destination_address_prefix: String IPv4 or IPv6 destination
            prefix and mask in the address/mask format. Example:
                '1.1.1.1'
                or
                '2001:db8::11/ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff'
        :return: StaticRoute object.
        """
        if isinstance(vrf, str):
            vrf = self.__get_vrf_from_switch(vrf)

        static_route_obj = self.session.api.get_module(
            self.session,
            "StaticRoute",
            destination_address_prefix,
            parent_vrf=vrf,
        )

        # Try to obtain data; if not, create
        try:
            static_route_obj.get()
        except GenericOperationError:
            # Create object inside switch
            static_route_obj.apply()

        return static_route_obj

    def static_nexthop(
        self,
        vrf,
        destination_address_prefix,
        next_hop_ip_address=None,
        nexthop_type=None,
        distance=None,
        next_hop_interface=None,
        bfd_enable=None,
    ):
        """
        Create a Static Nexthop, with a VRF and a Destination Address related
            to a Static Route.
        :param vrf: Name of the VRF on which the static route is to be
            configured. Defaults to "default". A Vrf object is also accepted.
        :param destination_address_prefix: String IPv4 or IPv6 destination
            prefix and mask in the address/mask format. A StaticRoute object
            is also accepted. Example:
                '1.1.1.1'
                or
                '2001:db8::11/ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff'
        :param next_hop_ip_address: The IPv4 address or the IPv6 address of
            next hop. Example:
                '1.1.1.1'
                or
                '2001:db8::11/ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff'
        :param nexthop_type: Specifies whether the static route is a forward,
            blackhole or reject route.
        :param distance: Administrative distance to be used for the next hop in
            the static route instead of default value.
        :param next_hop_interface: The interface through which the next hop can
            be reached.
        :param bfd_enable: Boolean to enable BFD.
        :return: StaticNexthop object.
        """
        if isinstance(vrf, str):
            vrf = self.__get_vrf_from_switch(vrf)
        static_route = destination_address_prefix
        if isinstance(destination_address_prefix, str):
            # Make a Static Route Object
            static_route_obj = self.session.api.get_module(
                self.session,
                "StaticRoute",
                destination_address_prefix,
                parent_vrf=vrf,
            )
            # Materialize Object to ensure its existence
            static_route_obj.get()
            static_route = static_route_obj

        if distance is None:
            distance = 1
        # Set variable
        next_hop_interface_obj = None
        if next_hop_interface is not None:
            next_hop_interface_obj = self.session.api.get_module(
                self.session, "Interface", next_hop_interface
            )
        if nexthop_type is None:
            nexthop_type = "forward"

        if nexthop_type == "forward":
            bfd_enable = False

        static_nexthop_obj = self.session.api.get_module(
            self.session,
            "StaticNexthop",
            0,
            parent_static_route=static_route,
        )

        # Try to obtain data; if not, create
        try:
            static_nexthop_obj.get()
            # Delete previous static nexthop
            static_nexthop_obj.delete()
        except GenericOperationError:
            # Catch error
            pass

        finally:
            static_nexthop_obj = self.session.api.get_module(
                self.session,
                "StaticNexthop",
                0,
                parent_static_route=static_route_obj,
                ip_address=next_hop_ip_address,
                distance=distance,
                port=next_hop_interface_obj,
                type=nexthop_type,
                bfd_enable=bfd_enable,
            )
            # Create object inside switch
            static_nexthop_obj.apply()

        return static_nexthop_obj

    def poe_interface(self, interface):
        """
        Create a PoE Interface object with associated settings.
        :param Interface: Alphanumeric name of the Interface the PoE_Interface
            belongs to. An Interface object is also accepted.
        :return: PoE Interface object.
        """
        if isinstance(interface, str):
            # Make Interface into an object
            interface_obj = self.session.api.get_module(
                self.session, "Interface", interface
            )
            # Materialize Interface to ensure its existence
            interface_obj.get()
            interface = interface_obj

        poe_interface_obj = self.session.api.get_module(
            self.session, "PoEInterface", interface
        )

        poe_interface_obj.get()

        return poe_interface_obj

    def mac(self, vlan, from_id, mac_address):
        """
        Create an Mac object.
        :param vlan_id: Numeric ID for VLAN. A Vlan object is also accepted.
        :param from_id: String source of the MAC address.
            Must be "dynamic", "VSX", "static", "VRRP", "port-access-security",
            "evpn", or "hsc".
        :param mac_address: String MAC address, or netaddr EUI object. Example:
            '01:02:03:04:05:06'.
        :return: Mac object.
        """
        if isinstance(vlan, int):
            vlan = self.vlan(vlan, "Vlan{0}".format(vlan))

        mac_obj = self.session.api.get_module(
            self.session,
            "Mac",
            from_id,
            mac_addr=mac_address,
            parent_vlan=vlan,
        )

        # Get MAC data
        mac_obj.get()

        return mac_obj

    def static_mac(self, vlan, port, mac_address):
        """
        Create an StaticMac object.
        param vlan_id: Numeric ID for VLAN. A Vlan object is also accepted.
        :param port: String for the Port's name. Example: 1/1/1.
        :param mac_address: String MAC address, or netaddr EUI object. Example:
            '01:02:03:04:05:06'.
        :return: StaticMac object.
        """
        if isinstance(vlan, int):
            vlan = self.vlan(vlan, "Vlan{0}".format(vlan))

        if isinstance(port, str):
            port = self.interface(port)

        static_mac_obj = self.session.api.get_module(
            self.session, "StaticMac", mac_address, parent_vlan=vlan, port=port
        )

        # Try to obtain data; if not, create
        try:
            static_mac_obj.get()
            if port is not None:
                if port.name != static_mac_obj.port.name:
                    static_mac_obj.port = port
                    static_mac_obj.apply()
        except GenericOperationError:
            # Create object inside switch
            static_mac_obj.apply()

        return static_mac_obj

    def qos(self, name, **kwargs):
        """
        Create a Qos object.
        :param name: String representing a user-defined name for a Qos object.
        :return: Returns a Qos object.
        """
        # Check for data type of name
        if not isinstance(name, str):
            raise ValueError("ERROR: Name must be on string format.")

        qos_obj = self.session.api.get_module(
            self.session, "Qos", name, **kwargs
        )

        # Try to obtain data; if unable to, create
        try:
            qos_obj.get()
            for k, v in kwargs.items():
                setattr(qos_obj, k, v)
        except GenericOperationError:
            pass  # not present in switch, not an error
        qos_obj.apply()
        return qos_obj

    def qos_cos(self, code_point, **kwargs):
        """
        Gets a QoS COS trust mode object.
        :param code_point: Integer to identify an entry a QoS COS trust mode
            object.
        :return: Returns a QoS COS trust mode object.
        """
        if not isinstance(code_point, int):
            raise ValueError("ERROR: Code Point must be an integer.")

        qos_cos_obj = self.session.api.get_module(
            self.session, "QosCos", code_point
        )

        # Try to obtain data only
        qos_cos_obj.get()

        # Review kwargs to change configurable attributes inside the object
        change_needed = False
        for attr, value in kwargs.items():
            if attr in qos_cos_obj.config_attrs:
                setattr(qos_cos_obj, attr, value)
                change_needed = True

        if change_needed:
            qos_cos_obj.apply()

        return qos_cos_obj

    def qos_dscp(self, code_point, **kwargs):
        """
        Retrieves a QoS DSCP trust mode map entry as an object.
        :param code_point: Integer to identify an entry a QoS DSCP trust mode
            object.
        :return: Returns a QoS DSCP trust mode object.
        """
        if not isinstance(code_point, int):
            raise Exception("ERROR: Code point must be an integer.")

        qos_dscp_obj = self.session.api.get_module(
            self.session, "QosDscp", code_point
        )

        # Try to obtain data only
        qos_dscp_obj.get()

        # Review kwargs to change configurable attributes inside the object
        change_needed = False
        for attr, value in kwargs.items():
            if attr in qos_dscp_obj.config_attrs:
                setattr(qos_dscp_obj, attr, value)
                change_needed = True

        if change_needed:
            qos_dscp_obj.apply()

        return qos_dscp_obj

    def queue(self, qos_name, queue_number, **kwargs):
        """
        Create a Queue object.
        :param qos_name: String with a user-defined name for a QoS object.
        :param queue_number: Integer representing a queue priority, with zero
            being the lowest priority. The maximum number of queues is
            hardware-dependent.
        :return: Queue object.
        """
        if not isinstance(qos_name, str):
            raise ValueError("ERROR: QoS name must be a string.")

        if not isinstance(queue_number, int):
            raise ValueError("ERROR: Queue number must be an integer.")

        queue_obj = self.session.api.get_module(
            self.session,
            "Queue",
            qos_name,
            queue_number=queue_number,
            **kwargs
        )

        # Try to obtain data; if unable to, create
        try:
            # Get the remote configuration, but the local one takes precedence
            queue_obj.get()
            for k, v in kwargs.items():
                setattr(queue_obj, k, v)
            queue_obj.apply()
        except GenericOperationError:
            queue_obj.create()

        # return object
        return queue_obj

    def queue_profile(self, name, **kwargs):
        """
        Create a Queue Profile object.
        :param name: name of the profile.
        :return: Queue Profile object.
        """
        profile = self.session.api.get_module(
            self.session, "QueueProfile", index_id=name, **kwargs
        )
        try:
            # Get the remote configuration, but the local one takes precedence
            profile.get()
            utils.create_attrs(profile, kwargs)  # so it gets re-applied here
        except GenericOperationError:
            pass
        finally:
            # Apply the local configuration to the switch
            profile.apply()
        return profile

    def queue_profile_entry(self, queue_number, queue_profile, **kwargs):
        """
        Create a Queue Profile Entry object.
        :param queue_number: Number that identifies the entry.
        :param queue_profile: A Queue Profile object.
        :return: Queue Profile Entry object.
        """
        if isinstance(queue_profile, str):
            queue_profile = self.session.api.get_module(
                self.session, "QueueProfile", index_id=queue_profile
            )

        entry = self.session.api.get_module(
            self.session,
            "QueueProfileEntry",
            index_id=queue_number,
            parent_profile=queue_profile,
            **kwargs
        )
        try:
            # Get the remote configuration, but the local one takes precedence
            entry.get()
            utils.create_attrs(entry, kwargs)  # so it gets re-applied here
        except GenericOperationError:
            pass
        finally:
            # Apply the local configuration to the switch
            entry.apply()
        return entry



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

from copy import deepcopy

import logging
import json
from urllib.parse import quote_plus

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError
from pyaoscx.exceptions.verification_error import VerificationError

from pyaoscx.utils import util as utils

from pyaoscx.pyaoscx_factory import PyaoscxFactory, Singleton
from pyaoscx.pyaoscx_module import PyaoscxModule


class Device(PyaoscxFactory, metaclass=Singleton):
    """
    Represents a Device and all of its attributes. Keeping all the important
        information inside one class.
    """

    base_uri = "system"

    def __init__(self, session):
        self.session = session
        self.firmware_version = None
        # Used to set attributes
        self.config_attrs = []
        self.materialized = False
        self.__original_attributes = {}
        # Set firmware version
        self.get_firmware_version()

    @PyaoscxModule.connected
    def get(self):
        """
        Perform a GET call to retrieve device attributes. After data from
            response, Device class attributes are generated using
            create_attrs().
        """
        logging.info("Retrieving the switch attributes and capabilities")

        non_configurable_attrs = [
            "admin_password_set",
            "aruba_central",
            "boot_time",
            "capabilities",
            "capacities",
            "mgmt_intf_status",
            "platform_name",
            "software_images",
            "software_info",
            "software_version",
            "qos_defaults",
        ]

        configurable_attrs = [
            "domain_name",
            "hostname",
            "other_config",
            "qos_config",
            "qos_default",
            "q_profile_default",
        ]

        # Concatenate both config and non-config attrs without duplicates
        all_attributes = list(set(non_configurable_attrs + configurable_attrs))

        attributes_list = ",".join(all_attributes)
        uri = "system?attributes={0}&depth={1}".format(
            attributes_list, self.session.api.default_depth
        )

        try:
            response = self.session.request("GET", uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        # Load into json format
        data = json.loads(response.text)

        # Create class attributes using util.create_attrs
        utils.create_attrs(self, data)

        utils.set_config_attrs(
            self, data, "config_attrs", non_configurable_attrs
        )

        # Save original attributes
        self.__original_attributes = deepcopy(
            utils.get_attrs(self, self.config_attrs)
        )
        # Set device as materialized
        self.materialized = True

    @property
    def modified(self):
        """
        Verifies if there has been a modification for this object or not.
        """
        device_data = utils.get_attrs(self, self.config_attrs)
        return device_data != self.__original_attributes

    @PyaoscxModule.connected
    def get_subsystems(self):
        """
        Perform GET call to retrieve subsystem attributes and create a
            dictionary containing them.
        """
        logging.info(
            "Retrieving the switch subsystem attributes and capabilities"
        )

        # Attribute list
        attributes = [
            "product_info",
            "power_supplies",
            "interfaces",
            "fans",
            "resource_utilization",
        ]

        # Format attribute list by joining every element with a comma
        attributes_list = ",".join(attributes)

        # Build URI
        uri = "system/subsystems?attributes={0}&depth={1}".format(
            attributes_list, self.session.api.default_subsystem_facts_depth
        )

        try:
            # Try to perform a GET call and retrieve the data
            response = self.session.request("GET", uri)

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        # Load into json format
        data = json.loads(response.text)
        data_subsystems = {"subsystems": data}

        # Create class attributes using util.create_attrs
        utils.create_attrs(self, data_subsystems)

    @PyaoscxModule.connected
    def get_firmware_version(self):
        """
        Perform a GET call to retrieve device firmware version.
        :return: firmware_version: The firmware version.
        """
        try:
            response = self.session.request("GET", "firmware")

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        self.firmware_version = data["current_version"]
        # Return Version
        return self.firmware_version

    @PyaoscxModule.materialized
    def apply(self):
        """
        Main method to update an existing Device object.
        :return modified: Boolean, True if object was created or modified.
        """
        return self.update()

    @PyaoscxModule.materialized
    def update(self):
        """
        Perform a PUT call to apply changes to a Device object.
        :return modified: Boolean, True if object was created or modified.
        """
        if not self.modified:
            return False

        device_data = utils.get_attrs(self, self.config_attrs)
        put_data = json.dumps(device_data)

        try:
            response = self.session.request("PUT", "system", data=put_data)
        except Exception as e:
            raise ResponseError("PUT", e)

        if not utils._response_ok(response, "PUT"):
            raise GenericOperationError(response.text, response.status_code)

        # Set new original attributes
        self.__original_attributes = deepcopy(device_data)

        return True

    ####################################################################
    # IMPERATIVE FUNCTIONS
    ####################################################################

    def update_banner(self, banner_info, banner_type="banner"):
        """
        Perform a PUT request to modify a Device's Banner.
        :param banner_info: String to be configured as the banner.
        :param banner_type: Type of banner being configured on the switch.
            Either banner or banner_exec.
        :return modified: Returns True if Banner was modified.
        """
        modified = False

        logging.info("Setting Banner")
        depth = self.session.api.default_depth

        # Second GET request to obtain just the variables that are writable
        selector = self.session.api.default_selector
        payload = {"depth": depth, "selector": selector}
        try:
            response = self.session.request(
                "GET", Device.base_uri, params=payload
            )

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        # Configurable data
        config_data = json.loads(response.text)

        # If Banner type does not exist
        if banner_type not in config_data["other_config"]:
            # Create Banner type
            config_data["other_config"][banner_type] = ""

        # Verify data is different
        if config_data["other_config"][banner_type] == banner_info:
            modified = False

        else:
            # Modify Banner
            config_data["other_config"][banner_type] = banner_info

            # UPDATE Banner
            put_uri = Device.base_uri
            # Set data to be used inside PUT
            put_data = json.dumps(config_data)

            try:
                response = self.session.request("PUT", put_uri, data=put_data)

            except Exception as e:
                raise ResponseError("PUT", e)

            if not utils._response_ok(response, "PUT"):
                raise GenericOperationError(
                    response.text, response.status_code, "UPDATE SYSTEM BANNER"
                )

            # Object was modified, returns True
            modified = True

        return modified

    def delete_banner(self, banner_type="banner"):
        """
        Perform a DELETE request to delete a device's Banner.
        :param banner_type: Type of banner being removed on the switch.
            Either banner or banner_exec.
        :return modified: Returns True if Banner was modified.
        """
        logging.info("Removing Banner")
        depth = self.session.api.default_depth

        # Second GET request to obtain just the variables that are writable
        selector = self.session.api.default_selector
        payload = {"depth": depth, "selector": selector}
        try:
            response = self.session.request(
                "GET", Device.base_uri, params=payload
            )

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        # Configurable data
        config_data = json.loads(response.text)

        # If Banner type does not exist
        if banner_type not in config_data["other_config"]:
            modified = False

        else:
            # Delete Banner
            config_data["other_config"].pop(banner_type)

            put_data = json.dumps(config_data)

            try:
                response = self.session.request(
                    "PUT", Device.base_uri, data=put_data
                )

            except Exception as e:
                raise ResponseError("PUT", e)

            if not utils._response_ok(response, "PUT"):
                raise GenericOperationError(
                    response.text, response.status_code, "DELETE Banner"
                )

            # Object was modified, returns True
            modified = True

        return modified

    def boot_firmware(self, partition_name="primary"):
        """
        Perform a POST request to Boot the AOS-CX switch with image present
            to the specified partition.
        :param partition_name: Name of the partition for device to boot to.
        :return bool: True if success.
        """
        # Lower case for partition name
        partition_name = partition_name.lower()
        if partition_name not in ["primary", "secondary"]:
            raise VerificationError("Boot Firmware", "Bad partition name")

        success = False
        uri = "boot?image={0}".format(partition_name)

        try:
            self.session.request("POST", uri)

        except Exception as e:
            raise ResponseError("POST", e)

        success = True

        # Return result
        return success

    def upload_firmware_http(
        self, remote_firmware_file_path, vrf, partition_name="primary"
    ):
        """
        Perform a PUT request to upload a firmware image given a http_request.
        :param remote_firmware_file_path: "HTTP server address and path for
            uploading firmware image, must be reachable through provided vrf
            ex: http://192.168.1.2:8000/TL_10_04_0030A.swi".
        :param vrf: VRF to be used to contact HTTP server, required if
            remote_firmware_file_path is provided.
        :param partition_name: Name of the partition for the image to be
            uploaded to.
        :return bool: True if success.
        """
        http_path = remote_firmware_file_path
        unsupported_versions = [
            "10.00",
            "10.01",
            "10.02",
            "10.03",
        ]
        # Verify Version
        for version in unsupported_versions:
            if version in self.firmware_version:
                raise VerificationError(
                    "Upload Firmware through HTTPs",
                    "Minimum supported firmware version is 10.04 for "
                    "remote firmware upload, your version is {0}".format(
                        self.firmware_version
                    ),
                )
        # Verify VRF
        if vrf is None:
            raise VerificationError(
                "VRF",
                "VRF needs to be provided in order"
                + " to upload firmware from HTTP server",
            )
        http_path_encoded = quote_plus(http_path)

        # Build URI
        uri = "firmware?image={0}&from={1}&vrf={2}".format(
            partition_name, http_path_encoded, vrf
        )

        # PUT for a HTTP Request
        try:
            response = self.session.request("PUT", uri)

        except Exception as e:
            raise ResponseError("PUT", e)

        if not utils._response_ok(response, "PUT"):
            raise GenericOperationError(response.text, response.status_code)

        # True if successful
        return True

    def upload_firmware_local(
        self, partition_name="primary", firmware_file_path=None
    ):
        """
        Perform a POST request to upload a firmware image from a local file.
        :param partition_name: Name of the partition for the image to be
            uploaded to.
        :param firmware_file_path: File name and path for local file uploading
            firmware image.
        :return success: True if success.
        """
        uri = "firmware?image={0}".format(partition_name)

        success = utils.file_upload(self.session, firmware_file_path, uri)
        # If no errors detected
        return success

    def upload_firmware(
        self,
        partition_name=None,
        firmware_file_path=None,
        remote_firmware_file_path=None,
        vrf=None,
    ):
        """
        Upload a firmware image from a local file OR from a remote location.
        :param partition_name: Name of the partition for the image to be
            uploaded to.
        :param firmware_file_path: File name and path for local file uploading
            firmware image.
            IMPORTANT: For this to be used, the remote_firmware_file_path
            parameter must be None.
        :param remote_firmware_file_path: HTTP server address and path for
            uploading firmware image, must be reachable through provided vrf
            ex: 'http://192.168.1.2:8000/TL_10_04_0030A.swi'.
        :param vrf: VRF to be used to contact HTTP server, required if
            remote_firmware_file_path is provided.
        :return bool: True if success.
        """
        result = None
        if partition_name is None:
            partition_name = "primary"

        # Use HTTP Server
        if remote_firmware_file_path is not None:
            result = self.upload_firmware_http(
                remote_firmware_file_path, vrf, partition_name
            )

        # Locally
        else:
            result = self.upload_firmware_local(
                partition_name, firmware_file_path
            )

        # If no errors detected
        return result

    def vsx_capable(self):
        """
        Return whether this device supports the VSX functionality.
        :return: True if device supports VSX.
        """
        return hasattr(self, "capabilities") and "vsx" in self.capabilities

    def is_capable(self, capability):
        """
        Check if the current Device has the given capability.
        :param capability: String name of a Device capability.
        :return: True if Device is capable.
        """
        if not self.materialized:
            self.get()

        return capability in self.capabilities



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

from abc import ABC, abstractmethod
from importlib import import_module

from pyaoscx.exceptions.parameter_error import ParameterError


class API(ABC):
    """
    Generic API class, that handles REST API versioning.
    """

    license = "Apache-2.0"

    def __str__(self):
        return self.version

    @classmethod
    def create(cls, target_version):
        """
        Translate the version string name to a valid python symbol.
        :param cls: API Class object.
        :param target_version: String with the API Version.
        :return api: API object.
        """
        version_name = "v" + target_version.replace(".", "_")

        try:
            # Import the appropriate API class based on the name.
            target_module = "pyaoscx.rest.{0}.api".format(version_name)
            api_class = getattr(import_module(target_module), version_name)

        except ModuleNotFoundError:
            raise ParameterError("Provided API version is not valid")

        return api_class()

    @classmethod
    @abstractmethod
    def __init__(self):
        """
        This method must be overwritten in the derived classes to set up the
            internal attributes, like version as minimum.
        """
        pass

    def valid_depth(self, depth):
        """
        Verifies if given depth is valid for the current API version.
        :param depth: Integer
        :return valid: Boolean True if depth is valid.
        """
        return depth in self.valid_depths

    def get_index(self, obj):
        """
        Method used to obtain the correct format of the objects information
            which depends on the Current API version.
        :param obj: PyaoscxModule object.
        :return info: Dictionary in the form of:
            "keepalive_vrf": {
                "keepalive_name": "Resource uri",
            }.
        """
        key_str = ""
        length = len(obj.indices)
        attributes = []
        for i in range(length):
            attr_name = obj.indices[i]
            attr_value = getattr(obj, attr_name)
            if not isinstance(attr_value, str):
                attr_value = str(attr_value)
            attributes.append(attr_value)

        key_str = ",".join(attributes)
        info = {key_str: obj.get_uri()}
        return info

    def get_keys(self, response_data, module_name=None):
        """
        Given a response_data obtain the indices of said dictionary and return
            them. Get keys should be used for only one element in the
            dictionary.
        :param response_data: a dictionary object in the form of:
            {
                "idx_1,idx_2": "/rest/v10.0X/system/<module>/<idx_1>,<idx_2>",
            }
        :return indices: List of indices.
        """
        indices = None
        for k, v in response_data.items():
            indices = k

        indices = indices.split(",")
        return indices

    def get_uri_from_data(self, data):
        """
        Given a response data, create a list of URI items. In this Version the
            data is a dict.
        :param data: Dictionary containing URI data in the form of:
            {
                '<name>': '/rest/v10.0X/system/<module>/<name>',
                '<name>': '/rest/v10.0X/system/<module>/<name>',
                '<name>': '/rest/v10.0X/system/<module>/<name>'
            }.
        :return uri_list: a list containing the input dictionary's values.
            Example:
            [
                '/rest/v10.0X/system/<module>/<name>',
                '/rest/v10.0X/system/<module>/<name>'
            ].
        """
        uri_list = []
        for k, v in data.items():
            uri_list.append(v)

        return uri_list

    def get_module_class(self, session, name):
        """
        Get a module's class given a session, and the module's name.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param name: Name of the module's class to be imported.
        """
        module_names = {
            "Interface": "rest.v{0}.interface".format(
                self.version.replace(".", "_")
            ),
            "Ipv6": "ipv6",
            "Vlan": "vlan",
            "Vrf": "vrf",
            "Vsx": "vsx",
            "BgpRouter": "bgp_router",
            "BgpNeighbor": "bgp_neighbor",
            "VrfAddressFamily": "vrf_address_family",
            "OspfRouter": "ospf_router",
            "Ospfv3Router": "ospfv3_router",
            "OspfVlink": "ospf_virtual_link",
            "OspfArea": "ospf_area",
            "OspfInterface": "ospf_interface",
            "DhcpRelay": "dhcp_relay",
            "ACL": "acl",
            "AclEntry": "acl_entry",
            "AggregateAddress": "aggregate_address",
            "StaticRoute": "static_route",
            "StaticNexthop": "static_nexthop",
            "PoEInterface": "poe_interface",
            "Mac": "mac",
            "StaticMac": "static_mac",
            "Qos": "qos",
            "QosCos": "qos_cos",
            "QosDscp": "qos_dscp",
            "Queue": "queue",
            "QueueProfile": "queue_profile",
            "QueueProfileEntry": "queue_profile_entry",
        }
        if name not in module_names:
            raise ParameterError(
                "Wrong module name. {0} doesn't exist".format(name)
            )
        return getattr(import_module("pyaoscx." + module_names[name]), name)

    def get_module(self, session, module, index_id=None, **kwargs):
        """
        Create a module object given a response data and the module's type.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device
        :param module: Name representing the module which is about to be
            created
        :param index_id: The module index_id or ID
        :return object: Return object same as module
        """
        module_class = self.get_module_class(session, module)

        if module == "OspfArea":
            return self._create_ospf_area(
                module_class, session, index_id, **kwargs
            )
        elif module == "Vsx":
            return self._create_vsx(module_class, session, **kwargs)
        else:
            return module_class(session, index_id, **kwargs)

    def _create_vsx(self, module_class, session, **kwargs):
        return module_class(session, **kwargs)



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging
import re

from urllib.parse import quote_plus, unquote_plus

from netaddr import EUI as MacAddress
from netaddr import mac_eui48

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError

from pyaoscx.utils import util as utils

from pyaoscx.pyaoscx_module import PyaoscxModule


class Mac(PyaoscxModule):
    """
    Provide configuration management for vlans/<id>/macs on AOS-CX devices.
    """

    indices = ["from", "mac_addr"]
    resource_uri_name = "macs"

    def __init__(self, session, from_id, mac_addr, parent_vlan, uri=None):
        """
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device
        :param from_id: String source of the MAC address. Must be "dynamic",
            "VSX", "static", "VRRP", "port-access-security", "evpn", or "hsc"
        :param mac_addr: String MAC address, or netaddr EUI object. Example:
            '01:02:03:04:05:06'.
        :param parent_vlan: Vlan object to which this MAC belongs.
        :param uri: Optional string containing the uri of the MAC object.
        """
        self.session = session

        # Set the MAC address format
        self.mac_format = mac_eui48
        self.mac_format.word_sep = ":"

        # Assign ID
        self.from_id = from_id
        self.mac_address = MacAddress(mac_addr, dialect=self.mac_format)
        # Assign parent VLAN
        self._set_vlan(parent_vlan)
        self._uri = uri
        # List used to determine attributes related to the MAC
        # configuration
        self.config_attrs = []
        self.materialized = False
        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self._original_attributes = {}
        # Attribute used to know if object was changed recently
        self.__modified = False
        self.display_name = "Mac"

        self.path = self._mac_path()

    def _set_vlan(self, parent_vlan):
        """
        Set parent Vlan object as an attribute for the MAC class. It is a
            private method because the user shouldn't be able to change the
            parent vlan.
        :param parent_vlan a Vlan object.
        """
        # Set parent VLAN
        self._parent_vlan = parent_vlan

        # Set URI
        self.base_uri = "{0}/{1}/{2}".format(
            self._parent_vlan.base_uri,
            self._parent_vlan.id,
            self.resource_uri_name,
        )

        macs = getattr(self._parent_vlan, self.resource_uri_name)
        found = False
        for mac in macs:
            if mac.mac_address == self.mac_address:
                found = True
                # Make list element point to current object
                mac = self
        if not found:
            macs.append(self)

    def _mac_path(self):
        """
        Get the path for internal purposes.
        """
        return "{0}/{1}{2}{3}".format(
            self.base_uri,
            self.from_id,
            self.session.api.compound_index_separator,
            quote_plus(str(self.mac_address)),
        )

    def _set_configuration_items(self, data, selector):
        # Determines if the MAC is configurable
        if selector in self.session.api.configurable_selectors:
            # Sets self.config_attrs and delete ID from it
            utils.set_config_attrs(
                self, data, "config_attrs", ["from", "mac_addr"]
            )

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a MAC table entry and fill
            the object with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if no exception is raised.
        """
        logging.info("Retrieving %s from switch", self)

        selector = selector or self.session.api.default_selector

        data = self._get_data(depth, selector)

        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        self._set_configuration_items(data, selector)

        # Set original attributes
        self._original_attributes = data
        # Remove both parts of the ID
        if "from" in self._original_attributes:
            self._original_attributes.pop("from")
        if "mac_addr" in self._original_attributes:
            self._original_attributes.pop("mac_addr")

        # Set port as an Interface Object
        if hasattr(self, "port") and self.port is not None:
            port_response = self.port
            # Instantiate empty object to use static method correctly
            interface_cls = self.session.api.get_module(
                self.session, "Interface", ""
            )
            # Set port as a Interface Object
            self.port = interface_cls.from_response(
                self.session, port_response
            )
            self.port.get()

        # Sets object as materialized
        # Information is loaded from the Device
        self.materialized = True
        return True

    @classmethod
    def get_all(cls, session, parent_vlan):
        """
        Perform a GET call to retrieve all system MACs inside a BGP Router, and
            create a dictionary containing them.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_vlan: parent bgp_router object where MAC is stored.
        :return: Dictionary containing MACs IDs as keys and a Mac objects as
            values.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)

        path = "{0}/{1}/{2}".format(
            parent_vlan.base_uri, parent_vlan.id, cls.resource_uri_name
        )

        try:
            response = session.request("GET", path)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        mac_dict = {}
        # Get all URI elements in the form of a list
        uri_list = session.api.get_uri_from_data(data)

        for uri in uri_list:
            # Create a Mac object
            indices, mac = cls.from_uri(session, parent_vlan, uri)
            mac_dict[indices] = mac

        return mac_dict

    @PyaoscxModule.connected
    def apply(self):
        """
        Not applicable for MAC.
        As HTTP Request method is not implemented inside REST API.
        """
        pass

    @PyaoscxModule.connected
    def update(self):
        """
        Not applicable for MAC.
        As HTTP Request method is not implemented inside REST API.
        """
        pass

    @PyaoscxModule.connected
    def create(self):
        """
        Not applicable for MAC.
        As HTTP Request method is not implemented inside REST API.
        """
        pass

    @PyaoscxModule.connected
    def delete(self):
        """
        Not applicable for MAC.
        As HTTP Request method is not implemented inside REST API.
        """
        pass

    @classmethod
    def from_response(cls, session, parent_vlan, response_data):
        """
        Create a Mac object given a response_data related to the MAC ID object.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_vlan: parent VLAN class where MAC is stored.
        :param response_data: The response must be a dictionary of the form:
            {
                "<id1>,<id2>": "/rest/v10.04/system/vlans/id/macs/id1,id2"
            }
        :return: Mac object.
        """
        mac_format = mac_eui48
        mac_format.word_sep = ":"
        mac_pair = session.api.get_keys(response_data, cls.resource_uri_name)
        mac_addr = mac_pair[1]
        from_id = mac_pair[0]

        mac_address = MacAddress(unquote_plus(mac_addr), dialect=mac_format)

        return Mac(session, from_id, mac_address, parent_vlan)

    @classmethod
    def from_uri(cls, session, parent_vlan, uri):
        """
        Create a Mac object.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_vlan: parent VLAN class where MAC is stored.
        :param uri: a String with a URI.
        :return indices, mac: tuple containing both the Mac object and
            the MAC' ID.
        """
        mac_format = mac_eui48
        mac_format.word_sep = ":"
        # Get ID from URI. Note that using a uri object here is of no
        # benefit because we are dealing with the path
        index_pattern = re.compile(
            r"(.*)macs/(?P<index1>.+)[,./-](?P<index2>.+)"
        )
        from_id = index_pattern.match(uri).group("index1")
        reference_mac_addr = index_pattern.match(uri).group("index2")

        mac_addr = MacAddress(
            unquote_plus(reference_mac_addr), dialect=mac_format
        )

        mac = Mac(session, from_id, mac_addr, parent_vlan)
        indices = "{0}{1}{2}".format(
            from_id, session.api.compound_index_separator, reference_mac_addr
        )

        return indices, mac

    def __str__(self):
        return str(self.mac_address)

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific MAC URI.
        return: Object's URI.
        """
        # TODO: remove this method in favor of uri_path once all
        # modules have been translated to the 'properties' style
        return self.uri_path

    @property
    def uri_path(self):
        """
        Method used to obtain the specific MAC URI.
        return: Object's URI.
        """
        if self._uri is None:
            self._uri = "{0}{1}/{2}{3}{4}".format(
                self.session.resource_prefix,
                self.base_uri,
                self.from_id,
                self.session.api.compound_index_separator,
                quote_plus(str(self.mac_address)),
            )

        return self._uri

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.
        return: Object format depending on the API Version.
        """
        # TODO: remove in favor of info_format when all modules are translated
        # to 'properties' style
        return self.info_format

    @property
    def info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.
        return: Object format depending on the API Version
        """
        return self.session.api.get_index(self)

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute
        :return: Boolean. True if the object was recently modified.
        """
        return self.modified

    ####################################################################
    # IMPERATIVE FUNCTIONS
    ####################################################################

    def get_info(self):
        """
        Perform a GET call to retrieve data for a MAC table entry and return
            info as a dictionary. Do not apply the configuration.
        :return info_dict: Returns a dictionary containing the current MAC
            Address information.
        """
        logging.info("Retrieving the switch MAC info")

        return self._get_data(None, None)



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging
import re

from urllib.parse import quote_plus, unquote_plus

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError

from pyaoscx.utils import util as utils
from pyaoscx.utils.list_attributes import ListDescriptor

from pyaoscx.static_nexthop import StaticNexthop

from pyaoscx.pyaoscx_module import PyaoscxModule


class StaticRoute(PyaoscxModule):
    """
    Provide configuration management for Static Route on AOS-CX devices.
    """

    indices = ["prefix"]
    static_nexthops = ListDescriptor("static_nexthops")
    resource_uri_name = "static_routes"

    def __init__(self, session, prefix, parent_vrf, uri=None, **kwargs):

        self.session = session
        # Assign IP
        self.prefix = unquote_plus(prefix)
        self.reference_address = quote_plus(self.prefix)
        self.address_family = utils.get_ip_version(self.prefix)
        # Assign parent Vrf object
        self.__set_vrf(parent_vrf)
        self._uri = uri
        # List used to determine attributes related to the Static Route
        # configuration
        self.config_attrs = []
        self.materialized = False
        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self.__original_attributes = {}
        # Set arguments needed for correct creation
        utils.set_creation_attrs(self, **kwargs)
        # Use to manage Ospf routers
        self.static_nexthops = []

        # Attribute used to know if object was changed recently
        self.__modified = False

    def __set_vrf(self, parent_vrf):
        """
        Set parent Vrf object as an attribute for the StaticRoute object.
        :param parent_vrf: a Vrf object.
        """
        # Set parent Vrf object
        self.__parent_vrf = parent_vrf

        # Set URI
        self.base_uri = "{0}/{1}/static_routes".format(
            self.__parent_vrf.base_uri,
            self.__parent_vrf.name,
        )

        # Verify Static Route doesn't exist already inside VRF
        for static_route in self.__parent_vrf.static_routes:
            if static_route.prefix == self.prefix:
                # Make list element point to current object
                static_route = self
            else:
                # Add self to static_routes list in parent Vrf object
                self.__parent_vrf.static_routes.append(self)

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a Static Route table entry and
            fill the object with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if there is not an exception raised.
        """
        logging.info("Retrieving %s from switch", self)

        depth = depth or self.session.api.default_depth
        selector = selector or self.session.api.default_selector

        if not self.session.api.valid_depth(depth):
            depths = self.session.api.valid_depths
            raise Exception("ERROR: Depth should be {0}".format(depths))

        if selector not in self.session.api.valid_selectors:
            selectors = " ".join(self.session.api.valid_selectors)
            raise Exception(
                "ERROR: Selector should be one of {0}".format(selectors)
            )

        payload = {"depth": depth, "selector": selector}

        uri = "{0}/{1}".format(self.base_uri, self.reference_address)
        try:
            response = self.session.request("GET", uri, params=payload)

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)
        # Delete unwanted data
        if "static_nexthops" in data:
            data.pop("static_nexthops")

        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        # Determines if the Static Route is configurable
        if selector in self.session.api.configurable_selectors:
            # Set self.config_attrs and delete ID from it
            utils.set_config_attrs(
                self, data, "config_attrs", ["prefix", "static_nexthops"]
            )

        # Set original attributes
        self.__original_attributes = data
        # Remove ID
        if "prefix" in self.__original_attributes:
            self.__original_attributes.pop("prefix")
        # Remove Static Nexthops
        if "static_nexthops" in self.__original_attributes:
            self.__original_attributes.pop("static_nexthops")

        # Clean Static Nexthops settings
        if self.static_nexthops == []:
            # Set Static Nexthops if any
            # Adds Static Nexthop to parent Vrf object
            StaticNexthop.get_all(self.session, self)

        # Sets object as materialized
        # Information is loaded from the Device
        self.materialized = True

        return True

    @classmethod
    def get_all(cls, session, parent_vrf):
        """
        Perform a GET call to retrieve all system Static Routes inside a VRF,
            and create a dictionary containing them.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_vrf: parent Vrf object where vrf is stored.
        :return: Dictionary containing Static Route IDs as keys and a Static
            Route objects as values.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)

        uri = "{0}/{1}/static_routes".format(
            parent_vrf.base_uri, parent_vrf.name
        )

        try:
            response = session.request("GET", uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        static_route_dict = {}
        # Get all URI elements in the form of a list
        uri_list = session.api.get_uri_from_data(data)

        for uri in uri_list:
            # Create a StaticRoute object and adds it to parent_vrf list
            prefix, static_route = StaticRoute.from_uri(
                session, parent_vrf, uri
            )
            # Load all Static Route data from within the Switch
            static_route.get()
            static_route_dict[prefix] = static_route

        return static_route_dict

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create a new or update an existing Static
            Route table entry. Checks whether the Static Route exists in the
            switch. Calls self.update() if object being updated. Calls
            self.create() if a new Static Route is being created.
        :return modified: Boolean, True if object was created or modified.
        """
        if not self.__parent_vrf.materialized:
            self.__parent_vrf.apply()

        modified = False
        if self.materialized:
            modified = self.update()
        else:
            modified = self.create()
        # Set internal attribute
        self.__modified = modified
        return modified

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing Static Route table
            entry.
        :return modified: True if Object was modified and a PUT request was
            made.
        """
        # Variable returned
        modified = False

        static_route_data = utils.get_attrs(self, self.config_attrs)

        uri = "{0}/{1}".format(self.base_uri, self.reference_address)

        # Compare dictionaries
        if static_route_data == self.__original_attributes:
            # Object was not modified
            modified = False

        else:
            post_data = json.dumps(static_route_data)

            try:
                response = self.session.request("PUT", uri, data=post_data)

            except Exception as e:
                raise ResponseError("PUT", e)

            if not utils._response_ok(response, "PUT"):
                raise GenericOperationError(
                    response.text, response.status_code
                )

            logging.info("SUCCESS: Updating %s", self)

            # Set new original attributes
            self.__original_attributes = static_route_data

            # Object was modified
            modified = True
        return modified

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new Static Route table entry. Only
            returns if no exception is raised.
        :return: Boolean, True if entry was created
        """
        # Add 'address_family' as a config attribute and remove duplicates
        self.config_attrs.append("address_family")
        self.config_attrs = list(set(self.config_attrs))

        static_route_data = utils.get_attrs(self, self.config_attrs)
        static_route_data["prefix"] = self.prefix
        static_route_data["vrf"] = self.__parent_vrf.get_uri()

        post_data = json.dumps(static_route_data)

        try:
            response = self.session.request(
                "POST", self.base_uri, data=post_data
            )

        except Exception as e:
            raise ResponseError("POST", e)

        if not utils._response_ok(response, "POST"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Adding %s", self)

        # Get all object's data
        self.get()
        # Object was created, thus modified
        return True

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform DELETE call to delete specified Static Route table entry.
        """
        uri = "{0}/{1}".format(self.base_uri, self.reference_address)

        try:
            response = self.session.request("DELETE", uri)

        except Exception as e:
            raise ResponseError("DELETE", e)

        if not utils._response_ok(response, "DELETE"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Deleting %s", self)

        # Delete back reference from VRF
        for static_route in self.__parent_vrf.static_routes:
            if static_route.prefix == self.prefix:
                self.__parent_vrf.static_routes.remove(static_route)

        # Delete object attributes
        utils.delete_attrs(self, self.config_attrs)

    @classmethod
    def from_response(cls, session, parent_vrf, response_data):
        """
        Create a StaticRoute object given a response_data related to the Static
            Route prefix object.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_vrf: parent Vrf object where Static Route is stored.
        :param response_data: The response must be a dictionary of the form:
            {
                prefix: "/rest/v10.04/system/vrfs/static_routes/prefix"
            }
        :return: Static Route Object.
        """
        static_route_arr = session.api.get_keys(
            response_data, StaticRoute.resource_uri_name
        )
        prefix = static_route_arr[0]
        return StaticRoute(session, prefix, parent_vrf)

    @classmethod
    def from_uri(cls, session, parent_vrf, uri):
        """
        Create a StaticRoute object given a URI.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_vrf: parent Vrf object where Static Route is stored.
        :param uri: a String with a URI.
        :return index, static_route: tuple containing both the static_route
            object and the static_route's prefix.
        """
        # Obtain ID from URI
        index_pattern = re.compile(r"(.*)static_routes/(?P<index>.+)")
        index = index_pattern.match(uri).group("index")

        # Create StaticRoute object
        static_route_obj = StaticRoute(session, index, parent_vrf, uri=uri)

        return index, static_route_obj

    def __str__(self):
        return "Static Route: {0}".format(self.prefix)

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific static route URI.
        return: Object's URI.
        """
        if self._uri is None:
            self._uri = "{0}{1}/{2}".format(
                self.session.resource_prefix,
                self.base_uri,
                self.reference_address,
            )

        return self._uri

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.
        return: Object format depending on the API Version.
        """
        return self.session.api.get_index(self)

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.
        :return: Boolean True if the object was recently modified.
        """
        return self.modified

    ####################################################################
    # IMPERATIVE FUNCTIONS
    ####################################################################

    def add_static_nexthop(
        self,
        next_hop_ip_address=None,
        next_hop_interface=None,
        distance=None,
        nexthop_type=None,
        bfd_enable=None,
    ):
        """
        Create a Static Nexthop, with a VRF and a Destination Address related
            to a Static Route.
        :param next_hop_ip_address: The IPv4 address or the IPv6 address of
            next hop.
            Example:
                '1.1.1.1'
                or
                '2001:db8::11/ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff'
        :param next_hop_interface: The interface through which the next hop
            can be reached.
        :param distance: Administrative distance to be used for the next
            hop in the static route instead of default value.
        :param nexthop_type: Specifies whether the static route is a forward,
            blackhole or reject route.
        :param bfd_enable: Boolean to enable BFD.
        :return: StaticNexthop object.
        """
        if distance is None:
            distance = 1
        # Set variable
        next_hop_interface_obj = None
        if next_hop_interface is not None:
            next_hop_interface_obj = self.session.api.get_module(
                self.session, "Interface", next_hop_interface
            )

        if nexthop_type is None:
            nexthop_type = "forward"

        if nexthop_type == "forward":
            bfd_enable = False

        static_nexthop_obj = self.session.api.get_module(
            self.session,
            "StaticNexthop",
            0,
            parent_static_route=self,
        )

        # Try to obtain data; if not, create
        try:
            static_nexthop_obj.get()
            # Delete previous static nexthop
            static_nexthop_obj.delete()
        except GenericOperationError:
            # Catch error
            pass

        finally:
            static_nexthop_obj = self.session.api.get_module(
                self.session,
                "StaticNexthop",
                0,
                parent_static_route=self,
                ip_address=next_hop_ip_address,
                distance=distance,
                port=next_hop_interface_obj,
                type=nexthop_type,
                bfd_enable=bfd_enable,
            )
            # Create object inside switch
            static_nexthop_obj.apply()

        return static_nexthop_obj



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging
import re

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError

from pyaoscx.utils import util as utils

from pyaoscx.pyaoscx_module import PyaoscxModule


class AclEntry(PyaoscxModule):
    """
    Provide configuration management for ACL Entry on AOS-CX devices.
    """

    indices = ["sequence_number"]
    resource_uri_name = "cfg_aces"

    protocol_dict = {
        "ah": 51,
        "esp": 50,
        "gre": 47,
        "icmp": 1,
        "icmpv6": 58,
        "igmp": 2,
        "ospf": 89,
        "pim": 103,
        "sctp": 132,
        "tcp": 6,
        "udp": 17,
    }

    # These parameters cannot be changed once the ACE is created
    # If any of them is set for update the entire ACE must be
    # replaced (delete and re-create). This list is needed while
    # we are not able to read the schema.
    immutable_parameter_names = [
        "action",
        "count",
        "dscp",
        "dst_ip",
        "dst_ip_group",
        "dst_l4_port_group",
        "dst_l4_port_max",
        "dst_l4_port_min",
        "dst_mac",
        "ecn",
        "ethertype",
        "fragment",
        "icmp_code",
        "icmp_type",
        "ip_precedence",
        "log",
        "pcp",
        "protocol",
        "sequence_number",
        "src_ip",
        "src_ip_group",
        "src_l4_port_group",
        "src_l4_port_max",
        "src_l4_port_min",
        "src_mac",
        "tcp_ack",
        "tcp_cwr",
        "tcp_ece",
        "tcp_established",
        "tcp_fin",
        "tcp_psh",
        "tcp_rst",
        "tcp_syn",
        "tcp_urg",
        "tos",
        "ttl",
        "vlan",
    ]

    # This list needs to be maintained only while we are not
    # able to read the schema. It is required when extracting
    # parameters to create copies.
    mutable_parameter_names = ["comment"]

    def __init__(
        self, session, sequence_number, parent_acl, uri=None, **kwargs
    ):

        self.session = session
        # Assign ID
        self.sequence_number = sequence_number
        # Assign parent Acl object
        self.__set_acl(parent_acl)
        self._uri = uri
        # List used to determine attributes related to the acl_entry
        # configuration
        self.config_attrs = []
        self.materialized = False
        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self.__original_attributes = {}
        # Set arguments needed for correct creation
        utils.set_creation_attrs(self, **kwargs)
        # Attribute used to know if object was changed recently
        self.__modified = False

    def __set_acl(self, parent_acl):
        """
        Set parent Acl object as an attribute for the AclEntry object.
        :param parent_acl: a Acl object.
        """
        # Set parent acl
        self.__parent_acl = parent_acl

        # Set URI
        self.base_uri = "{0}/{1}{2}{3}/cfg_aces".format(
            self.__parent_acl.base_uri,
            self.__parent_acl.name,
            self.session.api.compound_index_separator,
            self.__parent_acl.list_type,
        )

        # Verify acl_entry doesn't exist already inside acl
        for acl_entry in self.__parent_acl.cfg_aces:
            if acl_entry.sequence_number == self.sequence_number:
                # Make list element point to current object
                acl_entry = self
            else:
                # Add self to cfg_aces list in parent acl
                self.__parent_acl.cfg_aces.append(self)

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for an ACL Entry table entry and
            fill the object with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if no exception is raised.
        """
        logging.info("Retrieving %s from switch", self)

        depth = depth or self.session.api.default_depth
        selector = selector or self.session.api.default_selector

        if not self.session.api.valid_depth(depth):
            depths = self.session.api.valid_depths
            raise Exception("ERROR: Depth should be {0}".format(depths))

        if selector not in self.session.api.valid_selectors:
            selectors = " ".join(self.session.api.valid_selectors)
            raise Exception(
                "ERROR: Selector should be one of {0}".format(selectors)
            )

        payload = {"depth": depth, "selector": selector}

        uri = "{0}/{1}".format(self.base_uri, self.sequence_number)
        try:
            response = self.session.request("GET", uri, params=payload)

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        # Determines if the ACL Entry is configurable
        if selector in self.session.api.configurable_selectors:
            # Set self.config_attrs and delete ID from it
            utils.set_config_attrs(
                self, data, "config_attrs", ["sequence_number"]
            )

        # Set original attributes
        self.__original_attributes = data
        # Remove ID
        if "sequence_number" in self.__original_attributes:
            self.__original_attributes.pop("sequence_number")

        # Sets object as materialized
        # Information is loaded from the Device
        self.materialized = True
        return True

    @classmethod
    def get_all(cls, session, parent_acl):
        """
        Perform a GET call to retrieve all system ACL Entries inside an ACL,
            and create a dictionary containing them.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_acl: parent Acl object where ACL Entry is stored.
        :return acl_entry_dict: Dictionary containing ACL Entry IDs as keys
            and an ACL Entry objects as values.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)
        uri = "{0}/{1}{2}{3}/cfg_aces".format(
            parent_acl.base_uri,
            parent_acl.name,
            session.api.compound_index_separator,
            parent_acl.list_type,
        )

        try:
            response = session.request("GET", uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        acl_entry_dict = {}
        # Get all URI elements in the form of a list
        uri_list = session.api.get_uri_from_data(data)

        for uri in uri_list:
            # Create a AclEntry object and adds it to parent acl list
            sequence_number, acl_entry = AclEntry.from_uri(
                session, parent_acl, uri
            )
            # Load all acl_entry data from within the Switch
            acl_entry.get()
            acl_entry_dict[sequence_number] = acl_entry

        return acl_entry_dict

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create a new ACL Entry or update an existing
            one. It is possible that in case the are differences between the
            ACE on the switch and the local representation on immutable
            attributes a replace (delete+create) will take place. Note that
            unspecified parameters will be kept intact.
        :return modified: Boolean, True if object was created or modified.
        """
        if not self.__parent_acl.materialized:
            self.__parent_acl.apply()

        modified = False

        remote_ace = AclEntry(
            self.session, self.sequence_number, self.__parent_acl
        )

        try:
            # Should get all configurable attributes, not just the one that
            # are available for writing
            remote_ace.get(selector="configuration")
        except GenericOperationError:
            # If the get fails, the ACE doesn't exist, so a simple
            # create will suffice
            logging.info("%s for %s will be created", self, self.__parent_acl)
            modified = self.create()
        else:
            self._extract_missing_parameters_from(remote_ace)
            # Get was successful, so the ACE already exists
            if PyaoscxModule._is_replace_required(
                current=remote_ace,
                replacement=self,
                immutable_parameter_names=self.immutable_parameter_names,
            ):
                remote_ace.delete()
                logging.info(
                    "%s for %s will be replaced", self, self.__parent_acl.name
                )
                modified = self.create()
            else:
                # A replace was not required, so it is possible that an
                # update will suffice. Extracting the parameters from
                # the remote works as a materialization
                self.materialized = True
                modified = self.update()

        return modified

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing ACL Entry.
        :return modified: True if Object was modified and a PUT request
            was made.
        """
        # Variable returned
        modified = False

        acl_entry_data = utils.get_attrs(self, self.config_attrs)

        uri = "{0}/{1}".format(self.base_uri, self.sequence_number)

        if not self.__was_modified():
            # Object was not modified
            modified = False

        else:
            # Normally objects will pull from the switch only writable
            # attributes, but because of the replace (delete/create) behaviour
            # implemented for ACL Entries, we usually pull all configurable
            # attributes. Which means that hey have to be removed from the
            # request before being sent to the switch
            for key in list(acl_entry_data):
                if key in self.immutable_parameter_names:
                    del acl_entry_data[key]

            post_data = json.dumps(acl_entry_data)

            try:
                response = self.session.request("PUT", uri, data=post_data)

            except Exception as e:
                raise ResponseError("PUT", e)

            if not utils._response_ok(response, "PUT"):
                raise GenericOperationError(
                    response.text, response.status_code
                )

            logging.info("SUCCESS: Updating %s", self)
            # Set new original attributes
            self.__original_attributes = acl_entry_data

            # Object was modified
            modified = True

        if modified:
            self.__parent_acl._update_version()
            self.__parent_acl.apply()

        self.__modified = modified
        return modified

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new ACL Entry. Only returns if no
            exception is raised.
        :return modified: Boolean, True if entry was created
        """
        acl_entry_data = utils.get_attrs(self, self.config_attrs)
        acl_entry_data["sequence_number"] = self.sequence_number

        # Try to get protocol number
        try:
            if isinstance(self.protocol, str):
                if self.protocol == "any" or self.protocol == "":
                    acl_entry_data.pop("protocol")
                else:
                    protocol_num = self.protocol_dict[self.protocol]
                    acl_entry_data["protocol"] = protocol_num
            elif isinstance(self.protocol, int):
                acl_entry_data["protocol"] = self.protocol
        except Exception:
            pass
        post_data = json.dumps(acl_entry_data)

        try:
            response = self.session.request(
                "POST", self.base_uri, data=post_data
            )

        except Exception as e:
            raise ResponseError("POST", e)

        if not utils._response_ok(response, "POST"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Adding %s", self)

        # Get all object's data
        self.get()

        self.__parent_acl._update_version()
        self.__parent_acl.apply()

        # Object was created, means modified
        self.__modified = True
        return True

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform DELETE call to delete ACL Entry from parent ACL on the switch.
        """
        uri = "{0}/{1}".format(self.base_uri, self.sequence_number)

        try:
            response = self.session.request("DELETE", uri)

        except Exception as e:
            raise ResponseError("DELETE", e)

        if not utils._response_ok(response, "DELETE"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Deleting %s", self)

        # Delete back reference from ACL
        for acl_entry in self.__parent_acl.cfg_aces:
            if acl_entry.sequence_number == self.sequence_number:
                self.__parent_acl.cfg_aces.remove(acl_entry)

        # Delete object attributes
        utils.delete_attrs(self, self.config_attrs)

    @classmethod
    def from_response(cls, session, parent_acl, response_data):
        """
        Create a AclEntry object given a response_data related to the ACL Entry
            sequence_number object.
        :param cls: Class calling the method.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_acl: parent Acl object where ACL Entry is stored.
        :param response_data: The response must be a dictionary of the form:
            {
                <seq_number>: "/rest/v10.04/system/acls/cfg_aces/<seq_number>"
            }
        :return: AclEntry object.
        """
        acl_entry_arr = session.api.get_keys(
            response_data, AclEntry.resource_uri_name
        )
        sequence_number = acl_entry_arr[0]
        return AclEntry(session, sequence_number, parent_acl)

    @classmethod
    def from_uri(cls, session, parent_acl, uri):
        """
        Create a AclEntry object given a URI.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_acl: parent Acl object where ACL Entry is stored.
        :param uri: a String with a URI.
        :return index, acl_entry_obj: tuple containing both the AclEntry
            object and the acl_entry's sequence_number.
        """
        # Obtain ID from URI
        index_pattern = re.compile(r"(.*)cfg_aces/(?P<index>.+)")
        index = index_pattern.match(uri).group("index")

        # Create AclEntry object
        acl_entry_obj = AclEntry(session, index, parent_acl, uri=uri)

        return index, acl_entry_obj

    def __str__(self):
        return "ACL Entry ID {0}".format(self.sequence_number)

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific ACL Entry URI.
        return: AclEntry object's URI.
        """
        if self._uri is None:
            self._uri = "{0}{1}/{2}".format(
                self.session.resource_prefix,
                self.base_uri,
                self.sequence_number,
            )

        return self._uri

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.
        return: AclEntry object format depending on the API Version.
        """
        return self.session.api.get_index(self)

    @property
    def modified(self):
        return self.__modified

    def __was_modified(self):
        """
        Determine if the object was modified since the last materialization.
        """
        current = utils.get_attrs(self, self.config_attrs)
        original = self.__original_attributes
        if current == original:
            return False
        # Because ACL gets all configurable parameters and not just the
        # writable ones those two dictionaries could be different but because
        # on is missing members that the other one has. So it is necessary to
        # check that the missing members are set as None in the other.
        modified = False
        for key, value in current.items():
            if key in original:
                modified |= value != original[key]
            else:
                modified |= value is not None
        for key, value in original.items():
            if key in current:
                modified |= value != current[key]
            else:
                modified |= value is not None

        return modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.
        :return: Boolean True if the object was recently modified.
        """
        return self.modified

    ####################################################################
    # IMPERATIVE FUNCTIONS
    ####################################################################

    def modify(
        self,
        action=None,
        count=None,
        src_ip=None,
        dst_ip=None,
        dst_l4_port_min=None,
        dst_l4_port_max=None,
        src_mac=None,
        dst_mac=None,
        ethertype=None,
    ):
        """
        Create an AclEntry object, ACL Entry already exists, value passed won't
            update the entry.
        :param action: Action should be either "permit" or "deny".
        :param count: Optional boolean flag that when true, will make entry
            increment hit count for matched packets.
        :param src_ip: Optional source IP address. Both IPv4 and IPv6 are
            supported. Example:
                10.10.12.11/255.255.255.255
                2001:db8::11/ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff
        :param dst_ip: Optional destination IP address. Both IPv4 and IPv6 are
            supported. Example:
                10.10.12.11/255.255.255.255
                2001:db8::11/ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff
        :param dst_l4_port_min: Optional minimum L4 port number in range; used
            in conjunction with dst_l4_port_max.
        :param dst_l4_port_max: Optional maximum L4 port number in range; used
            in conjunction with dst_l4_port_min.
        :param src_mac: Optional source MAC address. Example:
            '01:02:03:04:05:06'
        :param dst_mac: Optional destination MAC address. Example:
            '01:02:03:04:05:06'
        :param ethertype: Optional integer EtherType number.
        :return: True if object was changed.
        """
        if action is not None:
            self.action = action

        if count is not None:
            self.count = count

        if src_ip is not None:
            self.src_ip = src_ip

        if dst_ip is not None:
            self.dst_ip = dst_ip

        if dst_l4_port_min is not None:
            self.dst_l4_port_min = dst_l4_port_min

        if dst_l4_port_max is not None:
            self.dst_l4_port_max = dst_l4_port_max

        if src_mac is not None:
            self.src_mac = src_mac

        if dst_mac is not None:
            self.dst_mac = dst_mac

        if ethertype is not None:
            self.ethertype = ethertype

        # Apply changes
        return self.apply()



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

from pyaoscx.exceptions.pyaoscx_error import PyaoscxError


class LoginError(PyaoscxError):
    """
    Exception raised for errors during login.
    """

    base_msg = "LOGIN ERROR"

    def __init__(self, *args):
        self.message = None
        self.status_code = None
        if args:
            msg = ", ".join((self.base_msg, str(args[0])))
            if len(args) > 1:
                self.status_code = args[1]
                msg = ", ".join(
                    str(a)
                    for a in (
                        msg,
                        *args[1:][1:],
                    )
                )
            self.message = msg
        else:
            self.message = self.base_msg



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging

from urllib.parse import quote_plus, unquote_plus

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError
from pyaoscx.exceptions.verification_error import VerificationError
from pyaoscx.utils import util as utils

from pyaoscx.pyaoscx_module import PyaoscxModule


class OspfInterface(PyaoscxModule):
    """
    Provide configuration management for OSPF Interface on AOS-CX devices.
    """

    collection_uri = (
        "system/vrfs/{name}/ospf{version}_routers/"
        "{instance_tag}/areas/{area_id}/ospf_interfaces"
    )
    object_uri = collection_uri + "/{interface_name}"

    indices = ["interface_name"]
    resource_uri_name = "ospf_interfaces"

    def __init__(self, session, interface_name, parent_ospf_area, **kwargs):
        self.session = session
        # Assign ID
        self.__interface_name = quote_plus(interface_name)
        # Assign parent OspfArea object
        self.__parent_ospf_area = parent_ospf_area
        port_name = kwargs.pop("port", interface_name)
        if port_name != interface_name:
            raise VerificationError(
                "OSPF interfaces must have the same name as the "
                "interface they are associated with."
            )
        self._set_port(interface_name)
        # List used to determine attributes related to the OSPF Interface
        # configuration
        self.config_attrs = []
        self.materialized = False
        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self._original_attributes = {}
        # Set arguments needed for correct creation
        utils.set_creation_attrs(self, **kwargs)
        # Attribute used to know if object was changed recently
        self.__modified = False
        self.base_uri = self.__parent_ospf_area.path + "/ospf_interfaces"
        self.path = "{0}/{1}".format(self.base_uri, quote_plus(interface_name))
        self.__parent_ospf_area.update_ospf_interfaces(self)

    @property
    def interface_name(self):
        return unquote_plus(self.__interface_name)

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @property
    def port(self):
        return self.interface_name

    def _set_port(self, name):
        from pyaoscx.interface import Interface

        self.__port = Interface(self.session, name)

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a OSPF Interfaces table entry
            and fill the object with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if there is not an exception raised.
        """
        logging.info("Retrieving %s from switch", self)
        selector = selector or self.session.api.default_selector
        data = self._get_data(depth, selector)
        if "port" in data:
            del data["port"]
        if "interface_name" in data:
            del data["interface_name"]
        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)
        # Determines if the OSPF Interfaces is configurable
        if selector in self.session.api.configurable_selectors:
            utils.set_config_attrs(self, data, "config_attrs")
        # Set original attributes
        self._original_attributes = data
        # Remove ID
        # Sets object as materialized
        # Information is loaded from the Device
        self.materialized = True
        return True

    @classmethod
    def get_all(cls, session, parent_ospf_area):
        """
        Perform a GET call to retrieve all system OSPF Interfaces inside a OSPF
            Area, and create a dictionary containing them as OspfInterface
            objects.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_ospf_area: parent OspfArea object where OspfInterface
            object is stored.
        :return: Dictionary containing OSPF Interface IDs as keys and a
            OspfInterface objects as values.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)
        uri = "{0}/{1}/ospf_interfaces".format(
            parent_ospf_area.base_uri, parent_ospf_area.area_id
        )
        try:
            response = session.request("GET", uri)
        except Exception as exc:
            raise ResponseError("GET", exc) from exc
        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)
        data = json.loads(response.text)
        ospf_interface_dict = {}
        # Get all URI elements in the form of a list
        uri_list = session.api.get_uri_from_data(data)
        for uri in uri_list:
            # Create a OspfInterface object
            interface_name, ospf_interface = cls.from_uri(
                session, parent_ospf_area, uri
            )
            # Load all OSPF Interfaces data from within the Switch
            ospf_interface_dict[interface_name] = ospf_interface
        return ospf_interface_dict

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create or update an existing OSPF Interface.
            Checks whether the OSPF Interface exists in the switch. Calls
            self.update() if OSPF Interface being updated. Calls self.create()
            if a new OSPF Interface is being created.
        :return modified: Boolean, True if object was created or modified.
        """
        if not self.__parent_ospf_area.materialized:
            self.__parent_ospf_area.apply()
        if self.materialized:
            return self.update()
        return self.create()

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing OSPF Interface.
        :return modified: True if Object was modified and a PUT request was
            made.
        """
        put_data = utils.get_attrs(self, self.config_attrs)
        # Get port uri
        put_data["port"] = self.__port.get_info_format()
        self.__modified = self._put_data(put_data)
        return self.__modified

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new OSPF Interface table entry. Only
            returns if an exception is not raised.
        :return: True if OSPF Interface table entry was added.
        """
        post_data = utils.get_attrs(self, self.config_attrs)
        post_data["port"] = self.__port.get_info_format()
        post_data["interface_name"] = self.interface_name

        self.__modified = self._post_data(post_data)
        return self.__modified

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform DELETE call to delete OSPF Interface.
        """
        self._send_data(self.path, None, "DELETE", "Delete")

        self.__parent_ospf_area.remove_ospf_interface(self)
        # Delete object attributes
        utils.delete_attrs(self, self.config_attrs)

    @classmethod
    def from_response(cls, session, parent_ospf_area, response_data):
        """
        Create a OspfInterface object given a response_data related to the OSPF
            Area ID object.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_ospf_area: parent OspfArea object where OspfInterface
            object is stored.
        :param response_data: The response must be a dictionary of the form:
            {id: URL}, with the URL being of the form:
            "ospf_routers/<id>/areas/<id>/ospf_interfaces/<interface_name>"
            under the path:
            "/rest/v10.04/system/vrfs/<vrf_name>/"
        :return: OspfInterface object.
        """
        ospf_interface_arr = session.api.get_keys(
            response_data, cls.resource_uri_name
        )
        ospf_interface_name = ospf_interface_arr[0]
        return cls(session, ospf_interface_name, parent_ospf_area)

    @classmethod
    def from_uri(cls, session, parent_ospf_area, uri):
        """
        Create a OspfInterface object given a URI.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :return interface_name, ospf_interface_obj: tuple containing both the
            OspfInterface name, and an OspfInterface object.
        """
        # Obtain ID from URI like:
        # system/vrfs/vrf/ospf_routers/1/areas/1.1.1.1/ospf_interfaces/iface
        interface_name = uri.split("/")[-1]

        # Create OspfInterface object
        ospf_interface_obj = cls(session, interface_name, parent_ospf_area)

        return interface_name, ospf_interface_obj

    def __str__(self):
        return "{0} with interface_name {1}".format(
            type(self).__name__, self.interface_name
        )

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific OSPF Interface uri.
        return: Object's URI.
        """
        return self.path

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.
        return: Object format depending on the API Version.
        """
        return self.session.api.get_index(self)

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.
        :return: Boolean True if the object was recently modified.
        """
        return self.modified



# (C) Copyright 2021-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError
from pyaoscx.exceptions.unsupported_capability_error import (
    UnsupportedCapabilityError,
)

from pyaoscx.utils import util as utils

from pyaoscx.device import Device
from pyaoscx.pyaoscx_module import PyaoscxModule

from pyaoscx.interface import Interface


class PoEInterface(Interface):
    """
    Provide configuration management for PoE Interface on AOS-CX devices.
    """

    resource_uri_name = "poe_interface"
    indices = ["name"]

    def __init__(self, session, parent_interface, uri=None, **kwargs):
        """
        Create an instance of PoEInterface Class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_interface: parent Inferface object where PoE is stored.
        :param uri: String with the PoE URI.
        """
        super(PoEInterface, self).__init__(session, parent_interface.name)

        self.session = session
        self._uri = uri
        # List used to determine attributes related to the PoEInterface
        # configuration
        self.config_attrs = []
        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self.__original_attributes = {}
        self.materialized = False
        # Set arguments needed for correct creation
        utils.set_creation_attrs(self, **kwargs)
        # Attribute used to know if object was changed recently
        self.__modified = False

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a PoE Interface table entry and
            fill the object with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if no exception is raised.
        """
        device = Device(self.session)
        if device.is_capable("quick_poe"):
            raise UnsupportedCapabilityError("This device is not PoE capable.")

        logging.info("Retrieving %s from switch", self)

        depth = depth or self.session.api.default_depth
        selector = selector or self.session.api.default_selector

        if not self.session.api.valid_depth(depth):
            depths = self.session.api.valid_depths
            raise ValueError("ERROR: Depth should be {0}".format(depths))

        if selector not in self.session.api.valid_selectors:
            selectors = " ".join(self.session.api.valid_selectors)
            raise ValueError(
                "ERROR: Selector should be one of {0}".format(selectors)
            )

        # Set payload
        payload = {"depth": depth, "selector": selector}

        # Set URI
        uri = "{0}/{1}/{2}".format(
            self.base_uri, self.percents_name, self.resource_uri_name
        )

        # Try to get a response to use its data
        try:
            response = self.session.request("GET", uri, params=payload)

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        # Determines if the PoE Interface is configurable
        if selector in self.session.api.configurable_selectors:
            # Set self.config_attrs
            utils.set_config_attrs(self, data, "config_attrs")

        # Set original attributes
        self.__original_attributes = data

        # Set object as materialized
        self.materialized = True

        return True

    @PyaoscxModule.connected
    def apply(self):
        """
        Apply an update of values of this PoE Interface.
        Calls self.update() to apply changes to an existing PoE Interface.
        """
        modified = False
        if self.materialized:
            modified = self.update()

        # Set internal attribute
        self.__modified = modified
        return modified

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing PoE Interface.
        :return modified: True if Object was modified and a PUT request was
            made.
        """
        # Variable returned
        modified = False

        poe_interface_data = utils.get_attrs(self, self.config_attrs)

        # Set URI
        uri = "{0}/{1}/{2}".format(
            self.base_uri, self.percents_name, self.resource_uri_name
        )

        # Compare dictionaries
        if poe_interface_data == self.__original_attributes:
            # Object was not modified
            modified = False

        else:
            post_data = json.dumps(poe_interface_data)
            try:
                response = self.session.request("PUT", uri, data=post_data)

            except Exception as e:
                raise ResponseError("PUT", e)

            if not utils._response_ok(response, "PUT"):
                raise GenericOperationError(
                    response.text, response.status_code
                )
            logging.info("SUCCESS: Updating %s", self)

            # Set new original attributes
            self.__original_attributes = poe_interface_data

            # Object was modified
            modified = True

        return modified

    @PyaoscxModule.connected
    def create(self):
        pass

    @PyaoscxModule.connected
    def delete(self):
        pass

    @PyaoscxModule.connected
    def get_all(self):
        pass

    @classmethod
    def from_uri(cls):
        pass

    def __str__(self):
        return "PoE Interface {0}".format(self.name)

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific PoE Interface URI.
        return: Object's URI.
        """
        if self._uri is None:
            self._uri = "{0}/{1}/{2}".format(
                self.base_uri, self.percents_name, self.resource_uri_name
            )

        return self._uri

    @PyaoscxModule.deprecated
    def get_info_format(self):
        pass

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.
        :return: Boolean True if the object was recently modified.
        """
        return self.modified

    ####################################################################
    # IMPERATIVE FUNCTIONS
    ####################################################################

    @PyaoscxModule.materialized
    def set_criticality(self, level):
        """
        Set the criticality level for the PoE Interface.
        :param level: String containing criticality level for the related
            PoE Interface. Valid criticality levels: 'low', 'high', and
            'critical'.
        :return: Returns True if there is not an exception raised.
        """
        valid_criticalities = ["low", "high", "critical"]
        if level not in valid_criticalities:
            raise ValueError(
                "ERROR: Criticality level must be one of {0}".format(
                    valid_criticalities
                )
            )

        # Set power level
        self.config["priority"] = level

        # Update changes
        return self.apply()

    @PyaoscxModule.materialized
    def set_power(self, state):
        """
        Perform a PUT call to set a configurable flag to control PoE power
            delivery on this Interface. A value of True would enable PoE power
            delivery on this Interface, and a value of False would disable PoE
            power delivery on this Interface.
        :return: True if Object was modified and a PUT request was made.
        """
        # Switches the state to a coherent value for the API documentation
        self.config["admin_disable"] = not state

        # Update changes
        return self.apply()



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging
import re

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError
from pyaoscx.exceptions.verification_error import VerificationError

from pyaoscx.utils import util as utils
from pyaoscx.utils.list_attributes import ListDescriptor

from pyaoscx.device import Device

from pyaoscx.pyaoscx_module import PyaoscxModule


class Vrf(PyaoscxModule):
    """
    Provide configuration management for VRF on AOS-CX devices.
    """

    base_uri = "system/vrfs"
    indices = ["name"]
    resource_uri_name = "vrfs"

    bgp_routers = ListDescriptor("bgp_routers")
    address_families = ListDescriptor("address_families")
    ospf_routers = ListDescriptor("ospf_routers")
    ospfv3_routers = ListDescriptor("ospfv3_routers")
    static_routes = ListDescriptor("static_routes")

    def __init__(self, session, name, uri=None, **kwargs):

        self.session = session
        self._uri = uri
        self.name = name
        # List used to determine attributes related to the VRF configuration
        self.config_attrs = []
        self.materialized = False
        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self.__original_attributes = {}
        # Set arguments needed for correct creation
        utils.set_creation_attrs(self, **kwargs)

        # Use to manage BGP Routers
        self.bgp_routers = []
        # Use to manage Vrf Address Families
        self.address_families = []
        # Use to manage OSPF Routers
        self.ospf_routers = []
        # Use to manage OSPFv3 Routers
        self.ospfv3_routers = []
        # Use to manage Static Routes
        self.static_routes = []
        # Attribute used to know if object was changed recently
        self.__modified = False

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a VRF table entry and fill the
            class with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information
            to return.
        :return: Returns True if no exception is raised.
        """
        logging.info("Retrieving %s from switch", self)

        depth = depth or self.session.api.default_depth
        selector = selector or self.session.api.default_selector

        if not self.session.api.valid_depth(depth):
            depths = self.session.api.valid_depths
            raise Exception("ERROR: Depth should be {0}".format(depths))

        if selector not in self.session.api.valid_selectors:
            selectors = " ".join(self.session.api.valid_selectors)
            raise Exception(
                "ERROR: Selector should be one of {0}".format(selectors)
            )

        payload = {"depth": depth, "selector": selector}

        uri = "{0}/{1}".format(Vrf.base_uri, self.name)

        try:
            response = self.session.request("GET", uri, params=payload)

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)
        # Delete unwanted data
        if "ospf_routers" in data:
            data.pop("ospf_routers")
        if "ospfv3_routers" in data:
            data.pop("ospfv3_routers")
        if "ospf_routers" in data or "ospfv3_routers" in data:
            data.pop("bgp_routers")
        if "static_routes" in data:
            data.pop("static_routes")

        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        # Determines if the VRF is configurable
        if selector in self.session.api.configurable_selectors:
            # Set self.config_attrs and delete ID from it
            unwanted_attributes = [
                "name",
                "type",
                "bgp_routers",
                "ospf_routers",
                "ospfv3_routers",
                "vrf_address_families",
                "static_routes",
            ]
            utils.set_config_attrs(
                self, data, "config_attrs", unwanted_attributes
            )
        # Set original attributes
        self.__original_attributes = data
        # Remove ID
        if "name" in self.__original_attributes:
            self.__original_attributes.pop("name")
        # Remove type
        if "type" in self.__original_attributes:
            self.__original_attributes.pop("type")
        # Remove bgp_routers
        if "bgp_routers" in self.__original_attributes:
            self.__original_attributes.pop("bgp_routers")
        # Remove ospf_routers
        if "ospf_routers" in self.__original_attributes:
            self.__original_attributes.pop("ospf_routers")
        # Remove static_routes
        if "static_routes" in self.__original_attributes:
            self.__original_attributes.pop("static_routes")
        # Remove vrf_address_families
        if "vrf_address_families" in self.__original_attributes:
            self.__original_attributes.pop("vrf_address_families")

        # Sets object as materialized
        # Information is loaded from the Device
        self.materialized = True

        device = Device(self.session)
        if not device.materialized:
            device.get()

        # Clean BGP Router settings
        if "bgp" in device.capabilities and self.bgp_routers == []:
            # gotta use deferred import to avoid cyclical import error
            from pyaoscx.bgp_router import BgpRouter

            # Set BGP Routers if any
            # Adds bgp_bouters to parent Vrf object
            BgpRouter.get_all(self.session, self)

        # Clean Address Families settings
        if self.address_families == []:
            # gotta use deferred import to avoid cyclical import error
            from pyaoscx.vrf_address_family import VrfAddressFamily

            # Set Address Families if any
            # Adds address_families to parent Vrf object
            VrfAddressFamily.get_all(self.session, self)

        # Clean OSPF Routers settings
        if "ospfv2" in device.capabilities and self.ospf_routers == []:
            # gotta use deferred import to avoid cyclical import error
            from pyaoscx.ospf_router import OspfRouter

            # Set OSPF Routers if any
            # Adds ospf_routers to parent Vrf object
            OspfRouter.get_all(self.session, self)

        # If no OSPFv3 Routers are present
        if "ospfv3" in device.capabilities and self.ospfv3_routers == []:
            # gotta use deferred import to avoid cyclical import error
            from pyaoscx.ospfv3_router import Ospfv3Router

            # Add all ospfv3_routers (if any) to parent Vrf object
            Ospfv3Router.get_all(self.session, self)

        # Clean Static Routess settings
        if self.static_routes == []:
            # gotta use deferred import to avoid cyclical import error
            from pyaoscx.static_route import StaticRoute

            # Set Static Route if any
            # Adds static_routes to parent Vrf object
            StaticRoute.get_all(self.session, self)
        return True

    @classmethod
    def get_all(cls, session):
        """
        Perform a GET call to retrieve all system VRFs and create a dictionary
            containing them.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :return: Dictionary containing VRF names as keys and a Vrf objects as
            values.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)

        try:
            response = session.request("GET", Vrf.base_uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        vrfs_dict = {}
        # Get all URI elements in the form of a list
        uri_list = session.api.get_uri_from_data(data)

        for uri in uri_list:
            # Create Vrf object
            name, vrf = Vrf.from_uri(session, uri)
            # Set VRF in dictionary
            vrfs_dict[name] = vrf

        return vrfs_dict

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create or update an existing VRF table
            entry. Checks whether the VRF exists in the switch. Calls
            self.update() if VRF is being updated. Calls self.create() if a new
            VRF is being created.
        :return modified: Boolean, True if object was created or modified.
        """
        modified = False
        if self.materialized:
            modified = self.update()
        else:
            modified = self.create()
        # Set internal attribute
        self.__modified = modified
        return modified

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing VRF table entry.
        :return modified: True if Object was modified and a PUT request was
            made.
        """
        vrf_data = utils.get_attrs(self, self.config_attrs)

        uri = "{0}/{1}".format(Vrf.base_uri, self.name)

        # Compare dictionaries
        # if vrf_data == self.__original_attributes:
        if json.dumps(vrf_data, sort_keys=True, indent=4) == json.dumps(
            self.__original_attributes, sort_keys=True, indent=4
        ):
            # Object was not modified
            modified = False

        else:
            post_data = json.dumps(vrf_data)

            try:
                response = self.session.request("PUT", uri, data=post_data)

            except Exception as e:
                raise ResponseError("PUT", e)

            if not utils._response_ok(response, "PUT"):
                raise GenericOperationError(
                    response.text, response.status_code
                )

            logging.info("SUCCESS: Adding %s", self)
            # Set new original attributes
            self.__original_attributes = vrf_data
            modified = True
        return modified

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new VRF using the object's attributes
            as POST body. Only returns if no exception is raised.
        :return modified: Boolean, True if entry was created
        """
        vrf_data = utils.get_attrs(self, self.config_attrs)

        vrf_data["name"] = self.name

        post_data = json.dumps(vrf_data)
        try:
            response = self.session.request(
                "POST", Vrf.base_uri, data=post_data
            )

        except Exception as e:
            raise ResponseError("POST", e)

        if not utils._response_ok(response, "POST"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Adding %s", self)

        # Get all objects data
        self.get()

        # Object was modified
        return True

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform DELETE call to delete VRF table entry.
        """
        # Delete object attributes
        utils.delete_attrs(self, self.config_attrs)

        uri = "{0}/{1}".format(Vrf.base_uri, self.name)

        try:
            response = self.session.request("DELETE", uri)

        except Exception as e:
            raise ResponseError("DELETE", e)

        if not utils._response_ok(response, "DELETE"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Deleting %s", self)

    @classmethod
    def from_response(cls, session, response_data):
        """
        Create a Vrf object given a response_data related to the Vrf object.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param response_data: The response must be a dictionary of the form:
            {
                "test_vrf": "/rest/v10.04/system/vrfs/test_vrf"
            }
        :return: Vrf object
        """
        # An interfaces's VRF is returned with an empty index in some
        # switch models (confirmed with a 4100), this empty index causes a
        # subsequent GET request to fail, because the constructed URI is
        # not correct, this may be due to these models supporting a single
        # VRF. This behavior occurs with the default VRF, so it's safe to
        # use 'default' when the value gotten in the response is the empty
        # string
        # TODO: determine if this is correct/intended behavior to take into
        # account for a possible refactor/rework to check for this earlier
        if next(iter(response_data)) == "":
            vrf_name = "default"
        else:
            vrf_name_arr = session.api.get_keys(
                response_data, cls.resource_uri_name
            )
            vrf_name = vrf_name_arr[0]
        return cls(session, vrf_name)

    @classmethod
    def from_uri(cls, session, uri):
        """
        Create a Vrf object given a VRF URI.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param uri: a String with a URI.
        :return name, vrf_obj: tuple containing both the VRF's name and a Vrf
            object.
        """
        # Obtain ID from URI
        index_pattern = re.compile(r"(.*)vrfs/(?P<index>.+)")
        name = index_pattern.match(uri).group("index")
        # Create vlan object
        vrf_obj = Vrf(session, name, uri=uri)

        return name, vrf_obj

    @classmethod
    def get_facts(cls, session):
        """
        Modify this to Perform a GET call to retrieve all VRFs and their
            respective data.
        :param cls: Class reference.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :return facts: Dictionary containing VRF IDs as keys and VRF objects as
            values.
        """
        logging.info("Retrieving switch VRF facts")

        # Set VRF facts depth
        vrf_depth = session.api.default_facts_depth

        # Build URI
        uri = "{0}?depth={1}".format(Vrf.base_uri, vrf_depth)

        try:
            # Try to get facts data via GET method
            response = session.request("GET", uri)

        except Exception as e:
            raise ResponseError("GET", e)
        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        # Load response text into json format
        facts = json.loads(response.text)

        return facts

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific VRF URI.
        return: Object's URI.
        """
        if self._uri is None:
            self._uri = "{0}{1}/{2}".format(
                self.session.resource_prefix,
                Vrf.base_uri,
                self.name,
            )

        return self._uri

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.
        return: Object format depending on the API Version.
        """
        return self.session.api.get_index(self)

    def __str__(self):
        return "VRF name: '{0}'".format(self.name)

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.
        :return: Boolean True if the object was recently modified.
        """
        return self.modified

    ####################################################################
    # IMPERATIVE FUNCTIONS
    ####################################################################

    def add_address_family(
        self, family_type="ipv4_unicast", export_target=[], import_targets=[]
    ):
        """
        Add a VRF Address Family to the current Vrf object.
        :param family_type: Alphanumeric type of the Address Family.
            The options are 'ipv4_unicast' and 'ipv6_unicast'.
            The default value is set to 'ipv4_unicast'.
        :param export_target: Optional list of export route targets.
        :param import_targets: Optional list of import route targets
        :return address_family: VrfAddressFamily Object
        """
        if not self.materialized:
            raise VerificationError(
                "VRF {0}".format(self.name), "Object not materialized"
            )

        # Verify if incoming address is a string
        if isinstance(family_type, str):
            # Create Vrf_Family_Address object -- add it to it's internal
            # address_families
            vrf_address_family = self.session.api.get_module(
                self.session,
                "VrfAddressFamily",
                family_type,
                parent_vrf=self,
                export_route_targets=export_target,
                import_route_targets=import_targets,
                route_map={},
            )
            # Try to get data, if non existent create
            try:
                # Try to obtain vrf_address_family address data
                vrf_address_family.get()
            # If vrf_address_family object is non existent, create it
            except GenericOperationError:
                # Create vrf_address_family inside switch
                vrf_address_family.apply()

        # Apply changes inside switch
        self.apply()

        return vrf_address_family

    def delete_address_family(self, family_type="ipv4_unicast"):
        """
        Given an address family type, delete that address from the current Vrf.
        :param family_type: Alphanumeric type of the Address Family.
            The options are 'ipv4_unicast' and 'ipv6_unicast'.
            A VrfAddressFamily object is accepted.
            The default value is set to 'ipv4_unicast'.
        """
        if not self.materialized:
            raise VerificationError(
                "VRF {0}".format(self.name), "Object not materialized"
            )
        # gotta use deferred import to avoid cyclical import error
        from pyaoscx.vrf_address_family import VrfAddressFamily

        # Verify if incoming address is a object
        if isinstance(family_type, VrfAddressFamily):
            # Obtain address
            family_type = family_type.address_family

        # Iterate through every address inside interface
        for add_family_obj in self.address_families:
            if add_family_obj.address_family == family_type:
                # Removing address does an internal delete
                self.address_families.remove(add_family_obj)

    def setup_dns(
        self,
        domain_name=None,
        domain_list=None,
        domain_servers=None,
        host_v4_address_mapping=None,
        host_v6_address_mapping=None,
    ):
        """
        Setup DNS client configuration within a VRF.
        :param domain_name: Domain name used for name resolution by the DNS
            client, if 'dns_domain_list' is not configured.
        :param domain_list: dict of DNS Domain list names to be used for
            address resolution, keyed by the resolution priority order.
            Example:
                {
                    0: "hpe.com"
                    1: "arubanetworks.com"
                }
        :param domain_servers: dict of DNS Name servers to be used for address
            resolution, keyed by the resolution priority order. Example:
                {
                    0: "4.4.4.10"
                    1: "4.4.4.12"
                }
        :param host_v4_address_mapping: dict of static host
            address configurations and the IPv4 address associated with them.
            Example:
                {
                    "host1": "5.5.44.5"
                    "host2": "2.2.44.2"
                }
        :param host_v6_address_mapping: dict of static host address
            configurations and the IPv6 address associated with them. Example:
                {
                    "host1": "2001:0db8:85a3:0000:0000:8a2e:0370:7334"
                }
        :return modified: Returns True if modified.
        """
        # Update Values

        if domain_name is not None:
            self.dns_domain_name = domain_name

        if domain_list is not None:
            self.dns_domain_list = domain_list

        if domain_servers is not None:
            self.dns_name_servers = domain_servers

        if host_v4_address_mapping is not None:
            self.dns_host_v4_address_mapping = host_v4_address_mapping

        if host_v6_address_mapping is not None:
            self.dns_host_v6_address_mapping = host_v6_address_mapping

        return self.apply()

    def delete_dns(
        self,
        domain_name=None,
        domain_list=None,
        domain_servers=None,
        host_v4_address_mapping=None,
        host_v6_address_mapping=None,
    ):
        """
        Delete DNS client configuration within a Vrf object.
        :param domain_name: If value is not None, it is deleted
        :param domain_list: If value is not None, it is deleted
        :param domain_servers: If value is not None, it is deleted
        :param host_v4_address_mapping: If value is not None, it is deleted
        :param host_v6_address_mapping: If value is not None, it is deleted
        :return modified: Returns True if modified.
        """
        # Update Values

        if domain_name is not None:
            self.dns_domain_name = None

        if domain_list is not None:
            self.dns_domain_list = None

        if domain_servers is not None:
            self.dns_name_servers = None

        if host_v4_address_mapping is not None:
            self.dns_host_v4_address_mapping = None

        if host_v6_address_mapping is not None:
            self.dns_host_v6_address_mapping = None

        return self.apply()

    def update_ospf_routers(self, router):
        """
        Update references to OSPF Routers. If a Router with the same instance
            tag is found, update the reference to the new router, otherwise,
            add reference to the new router.
        """
        routers = getattr(self, router.resource_uri_name)
        for r in routers:
            if r.instance_tag == router.instance_tag:
                # Make list element point to current object
                # See utils.list_attributes.ListDescriptor
                r = router
                return
        routers.append(router)

    def remove_ospf_router(self, router):
        """
        Update references to OSPF Routers. If a Router with the same instance
            tag is found, delete the reference to it.
        """
        routers = getattr(self, router.resource_uri_name)
        for r in routers:
            if r.instance_tag == router.instance_tag:
                routers.remove(r)
